import argparse
import enum
import itertools
import os
import os.path
import socket
from pathlib import Path
import ssl
import sys
import tempfile
from typing import Optional, Dict, List, NamedTuple, Tuple, Iterable, Set, Counter, TypeVar
import urllib.error
import urllib.parse
import urllib.request

from .steps import UpgradeStep, UpgradeAppliancesStep, InstallAppliancesUpgradeStep, \
    DisableExtraControllersStep, EnableExtraControllersStep, CancelAppliancesUpgradeStep, _BaseUpgradeAppliancesStep, \
    PrepareAppliancesUpgradeStep, ImageSpecifier, ImageFilePath, UploadUpgradeImageToControllerStep, ImageURL, \
    DownloadUpgradeImageToControllerStep, LazyImageURL, DeleteUpgradeImageFromControllerStep
from .utils import plural, ask_for_confirmation, default_device_id
from .admin import Admin, SSLError, AdminError, URLConnectionError, HTTPError, URLError, FileStatus, ProgressCallback, \
    APIVersionError
from .appliance import ApplianceAdmin, Appliance, get_appliances_status, log_status_changes, \
    ApplianceStatus, guess_target_version, APPLIANCE_VERSION_REGEXP, ApplianceRole, \
    APPLIANCE_SEMVER_REGEXP, UpgradeStatus, parse_server, DetailedUpgradeStatus, \
    UpgradeInProgressError, UpgradeError, build_appliance_filter, APPLIANCE_FILTER_ATTRIBUTES, \
    MINIMUM_VERSION_HA_PATCH_CONTROLLER_UPGRADES, get_appliances_upgrade_status, AppliancesStartingStatus, names, \
    MINIMUM_PEER_API_VERSION, DEFAULT_PREPARE_TIMEOUT, DEFAULT_COMPLETE_TIMEOUT, MINIMUM_PREPARE_TIMEOUT, \
    MINIMUM_COMPLETE_TIMEOUT, DEFAULT_THROTTLE, TIMEOUTS
from .logger import log, LOG_LEVEL_MAPPING, pick_color_function, set_log_level

# Version of the Appgate SDP appliances upgrade script
VERSION = '5.3.3'

T = TypeVar('T')

HELP_PREFIX = 'https://sdphelp.appgate.com/adminguide/v5.3'


class CriticalUpgradeError(UpgradeError):
    """
    The upgrade failed and the initial state could not be restored!
    """
    pass


class UpgradeCancelled(Exception):
    pass


UpgradePlan = NamedTuple('UpgradePlan', [
    ('master_controller', Optional[Appliance]),
    ('controllers', List[Appliance]),
    ('extra_controllers', List[Appliance]),
    ('all_controllers', List[Appliance]),
    ('log_servers', List[Appliance]),
    ('other_groups', List[List[Appliance]]),
])


def appliance_group_id(appliance: Appliance) -> str:
    """
    Return a group id used to group appliances logically, e.g. all gateways
    belonging to the same site. Use hashlib instead of the built-in hash
    to make the output stable across runs.
    """
    import hashlib
    m = hashlib.sha256()
    if appliance.site:
        m.update(appliance.site.encode('utf-8'))
    if appliance.log_forwarder_sites:
        # We want a single log forwarder site to result in a different
        # group id than a single gateway site so they get grouped separately.
        m.update(str(appliance.log_forwarder_sites).encode('utf-8'))
    for role in {ApplianceRole.GATEWAY, ApplianceRole.CONNECTOR}:
        if role in appliance.roles:
            # Group gateways and connectors separately to make sure we don't
            # upgrade e.g. all gateways at the same time.
            m.update(role.value.encode('utf-8'))
    group_id = m.hexdigest()
    return group_id


def list_split(l: List[T], n: int) -> List[List[T]]:
    """
    Splits a list l in parts.

    The resulting list of lists contains all the elements.
    The content is not too well sorted as the remaining elements
    are added at the end.
    """
    slice = len(l) // n

    r = []

    for i in range(n):
        r.append(l[i*slice:i*slice + slice])

    # Add the remainder
    if len(l) % n:
        for index, v in enumerate(l[-(len(l) % n):]):
            r[index].append(v)
    return r


def split_appliances_by_group(appliances: List[Appliance], site_steps: int) -> List[List[Appliance]]:
    """
    Split appliances in two upgrade groups, by taking half from each group,
    one upgrade group will be upgraded at a time to minimize downtime.
    """
    # Group the appliances by sites.
    appliances_by_group = {}  # type: Dict[str, List[Appliance]]
    for appliance in appliances:
        group_id = appliance_group_id(appliance)
        if group_id not in appliances_by_group:
            appliances_by_group[group_id] = []
        appliances_by_group[group_id].append(appliance)

    groups = [list() for _ in range(site_steps)]  # type: List[List[Appliance]]

    for i in appliances_by_group.values():
        for index, value in enumerate(list_split(i, site_steps)):
            groups[index].extend(value)

    return groups


def prepare_upgrade_plan(master_controller: Appliance, appliances: List[Appliance],
                         all_controllers: List[Appliance], site_steps: int) -> UpgradePlan:
    """
    Create a structure representing an upgrade plan from the list of all appliances.
    appliances is the list of all appliances to upgrade.
    all_controllers is the list of all controllers including those that are not getting upgraded.
    """
    controllers = [
        appliance for appliance in appliances
        if ApplianceRole.CONTROLLER in appliance.roles
    ]

    # Filter the master controller we are connecting to...
    extra_controllers = [
        controller for controller in controllers
        if controller.appliance_id != master_controller.appliance_id
    ]

    log_servers = [
        appliance for appliance in appliances
        if ApplianceRole.LOGSERVER in appliance.roles and ApplianceRole.CONTROLLER not in appliance.roles
    ]

    other_appliances = [
        appliance for appliance in appliances
        if appliance.roles.isdisjoint({ApplianceRole.CONTROLLER, ApplianceRole.LOGSERVER})
    ]

    # Group remaining appliances to avoid wasting maintenance window time
    other_groups = split_appliances_by_group(other_appliances, site_steps)

    # Don't give the master controller to the upgrade plan if it doesn't need upgrading
    master_controller_to_upgrade = None  # type: Optional[Appliance]
    if master_controller in appliances:
        master_controller_to_upgrade = master_controller

    return UpgradePlan(
        master_controller=master_controller_to_upgrade,
        controllers=controllers,
        extra_controllers=extra_controllers,
        all_controllers=all_controllers,
        log_servers=log_servers,
        other_groups=other_groups,
    )


def ask_for_upgrade_confirmation(yes: bool) -> None:
    """
    Will raise an upgrade cancelled exception if the user did not confirm.
    """
    question = 'Do you want to continue'
    if not ask_for_confirmation(question, yes):
        raise UpgradeCancelled('Cancelled by user!')
    return None


def appliance_hostnames(appliance: Appliance) -> Set[str]:
    """
    Return the set of normalized appliance peer and admin hostnames.
    """
    hostnames = set()  # type: Set[str]
    hostnames.add(appliance.peer_hostname.casefold())
    if appliance.admin_hostname:
        hostnames.add(appliance.admin_hostname.casefold())
    return hostnames


def appliances_hostnames(appliances: Iterable[Appliance]) -> Set[str]:
    """
    Return the set of normalized appliances peer and admin hostnames.
    """
    hostnames = set()  # type: Set[str]
    for appliance in appliances:
        hostnames.update(appliance_hostnames(appliance))
    return hostnames


def count_appliance_hostnames(appliances: List[Appliance]) -> Counter[str]:
    """
    Count how many appliances use each peer and admin hostnames.
    """
    counter = Counter[str]()
    for appliance in appliances:
        counter.update(appliance_hostnames(appliance))
    return counter


def validate_hostname_uniqueness(hostname: str, port: int) -> None:
    """
    Validate that the given hostname resolves to at most 1 ip per ip version.
    """
    records = socket.getaddrinfo(hostname, port, type=socket.SOCK_STREAM)
    ips_v4 = {
        result[0]
        for family, _, _, _, result in records
        if family == socket.AF_INET
    }
    ips_v6 = {
        result[0]
        for family, _, _, _, result in records
        if family == socket.AF_INET6
    }
    if len(ips_v4) > 1 or len(ips_v6) > 1:
        raise UpgradeError(
            'The given hostname {!r} does not resolve to a unique ip. '
            'A unique ip is necessary to ensure that the upgrade script '
            'connects to only the (primary) administration Controller. '
            'The hostname resolves to the following ips: {}'.format(
                hostname, ', '.join(sorted(ips_v4 | ips_v6))))


def find_master_controller(hostname: str, appliances: List[Appliance]) -> Appliance:
    """
    The given hostname should match one of the controller's actual admin hostname.

    Hostnames should be compared in a case insensitive way.
    """
    controllers = [
        appliance for appliance in appliances
        if ApplianceRole.CONTROLLER in appliance.roles
    ]
    hostname_counter = count_appliance_hostnames(appliances)
    candidate_hostnames = {
        controller_hostname for controller_hostname in appliances_hostnames(controllers)
        if hostname_counter[controller_hostname] == 1
    }

    if hostname_counter[hostname.casefold()] > 1:
        raise UpgradeError(
            'The given controller hostname {} is used by more than one appliance. '
            'A unique controller admin (or peer) hostname is required to perform the upgrade. '
            'Use one of the following controller admin (or peer) hostnames: {}'.format(
                hostname, ', '.join(sorted(candidate_hostnames))))

    try:
        return next(controller for controller in controllers
                                 if hostname.casefold() in appliance_hostnames(controller))
    except StopIteration:
        raise UpgradeError(
            'Unable to match the given controller hostname {} with the actual controller admin (or peer) hostname, '
            'use the --actual-hostname option to identify this controller\'s actual admin (or peer) hostname: {}'.format(
                hostname, ', '.join(sorted(candidate_hostnames)))) from None


def prepare_image_url(admin: ApplianceAdmin, master_controller: Appliance,
                      image: ImageSpecifier, yes: bool) -> Tuple[LazyImageURL, List[UpgradeStep]]:
    upgrade_steps = []  # type: List[UpgradeStep]
    if isinstance(image, ImageFilePath):
        image_url = LazyImageURL()
        upgrade_steps.append(UploadUpgradeImageToControllerStep(
            admin=admin,
            image_path=image.image_path,
            image_url=image_url,
            peer_hostname=master_controller.peer_hostname,
            peer_port=master_controller.peer_port,
            yes=yes
        ))
    elif isinstance(image, ImageURL):
        if image.host_on_controller:
            image_url = LazyImageURL()
            upgrade_steps.append(DownloadUpgradeImageToControllerStep(
                admin=admin,
                source_image_url=image.image_url,
                image_url=image_url,
                peer_hostname=master_controller.peer_hostname,
                peer_port=master_controller.peer_port,
                yes=yes
            ))
        else:
            image_url = LazyImageURL(image.image_url)
    else:
        raise UpgradeError('Invalid image specifier: {}'.format(image))
    return image_url, upgrade_steps


def prepare_prepared_upgrade_steps(admin: ApplianceAdmin, appliances: List[Appliance],
                                   master_controller: Appliance,
                                   image: ImageSpecifier,
                                   dev_keyring: bool,
                                   yes: bool = False,
                                   throttle: Optional[int] = None) -> List[UpgradeStep]:
    image_url, upgrade_steps = prepare_image_url(admin, master_controller, image, yes)

    upgrade_steps.append(PrepareAppliancesUpgradeStep(
        admin=admin,
        appliances=appliances,
        image_url=image_url,
        throttle=throttle,
        dev_keyring=dev_keyring,
    ))

    # Upgrade image is not set, that means that we will get it to the controller
    # first, make sure we remove it as well.
    if image_url and image_url.value is None:
        upgrade_steps.append(DeleteUpgradeImageFromControllerStep(
            admin=admin,
            image_url=image_url
        ))
    return upgrade_steps


def prepare_cancel_upgrade_steps(admin: ApplianceAdmin, appliances: List[Appliance],
                                 delete_image: Optional[str] = None) -> List[UpgradeStep]:
    upgrade_steps = []  # type: List[UpgradeStep]
    if len(appliances) > 0:
        upgrade_steps.append(CancelAppliancesUpgradeStep(
            admin=admin,
            appliances=appliances
        ))
    if delete_image is not None:
        upgrade_steps.append(DeleteUpgradeImageFromControllerStep(
            admin=admin,
            image_url=LazyImageURL(delete_image)
        ))
    return upgrade_steps


def _upgrade_appliances_step(admin: ApplianceAdmin, image_url: Optional[LazyImageURL], dev_keyring: bool,
                             appliances: List[Appliance], starting_status: AppliancesStartingStatus,
                             appliance_description: Optional[str] = None, expected_state: Optional[str] = None,
                             check_status: bool = True, throttle: Optional[int] = None) -> _BaseUpgradeAppliancesStep:
    """
    Return an install appliances upgrade step if image url is None, else return an upgrade appliances step.
    """
    if image_url is None:
        return InstallAppliancesUpgradeStep(
            admin=admin, appliances=appliances, starting_status=starting_status,
            appliance_description=appliance_description, expected_state=expected_state,
            check_status=check_status, throttle=throttle
        )
    else:
        return UpgradeAppliancesStep(
            admin=admin, appliances=appliances, starting_status=starting_status,
            dev_keyring=dev_keyring,
            appliance_description=appliance_description, expected_state=expected_state,
            check_status=check_status, image_url=image_url, throttle=throttle
        )


def prepare_upgrade_steps(admin: ApplianceAdmin, upgrade_plan: UpgradePlan, master_controller: Appliance,
                          image: Optional[ImageSpecifier],
                          dev_keyring: bool,
                          starting_status: AppliancesStartingStatus, yes: bool = False,
                          throttle: Optional[int] = None, patch: bool = False) -> List[UpgradeStep]:
    image_url = None  # type: Optional[LazyImageURL]
    upgrade_steps = []  # type: List[UpgradeStep]
    if image is not None:
        image_url, upgrade_steps = prepare_image_url(admin, master_controller, image, yes)

    # Take into consideration the controllers that are not being upgraded when
    # disabling and enabling the controller role if upgrading at least one controller.
    controllers_but_master = None  # type: Optional[List[Appliance]]
    if upgrade_plan.master_controller:
        controllers_but_master = [
            appliance for appliance in upgrade_plan.all_controllers
            if appliance.appliance_id != upgrade_plan.master_controller.appliance_id
        ]

    # Disable all but one controller, unnecessary for patch upgrades
    if not patch and controllers_but_master is not None and len(controllers_but_master) > 0:
        disable_extra_controllers = DisableExtraControllersStep(
            admin=admin,
            master_controller=upgrade_plan.master_controller,
            extra_controllers=controllers_but_master
        )
        upgrade_steps.append(disable_extra_controllers)

    # Upgrade master controller
    if upgrade_plan.master_controller is not None:
        if patch:
            # The controller should go back to it's original state
            expected_state = starting_status.status[upgrade_plan.master_controller.name].state
            assert expected_state
        else:
            expected_state = 'single_controller_ready'
        upgrade_controllers = _upgrade_appliances_step(
            admin=admin,
            image_url=image_url,
            dev_keyring=dev_keyring,
            appliances=[upgrade_plan.master_controller],
            starting_status=starting_status,
            appliance_description='first controller',
            expected_state=expected_state,
            # Can't check status when the controller is down
            check_status=False,
            throttle=throttle
        )
        upgrade_steps.append(upgrade_controllers)

    # Upgrade extra controllers
    if len(upgrade_plan.extra_controllers) > 0:
        if patch:
            # Extra controllers should be upgraded one by one if they are active controllers.
            if image_url:
                upgrade_steps.append(PrepareAppliancesUpgradeStep(
                    admin=admin,
                    appliances=upgrade_plan.extra_controllers,
                    image_url=image_url,
                    throttle=throttle,
                    dev_keyring=dev_keyring,
                ))
            # Support up to 6 controllers, use explicit names for those.
            ordinals = ['second', 'third', 'fourth', 'fifth', 'sixth'][:len(upgrade_plan.extra_controllers)]
            for extra_controller, ordinal in itertools.zip_longest(
                    upgrade_plan.extra_controllers, ordinals, fillvalue='additional'):
                upgrade_steps.append(_upgrade_appliances_step(
                    admin=admin,
                    image_url=None,
                    dev_keyring=dev_keyring,
                    appliances=[extra_controller],
                    starting_status=starting_status,
                    expected_state='multi_controller_ready',
                    appliance_description='{} controller'.format(ordinal),
                    throttle=throttle
                ))
        else:
            expected_state = 'appliance_ready'
            upgrade_extra_controllers = _upgrade_appliances_step(
                admin=admin,
                image_url=image_url,
                dev_keyring=dev_keyring,
                appliances=upgrade_plan.extra_controllers,
                starting_status=starting_status,
                expected_state=expected_state,
                appliance_description='additional controller',
                throttle=throttle
            )
            upgrade_steps.append(upgrade_extra_controllers)

    # Enable all controllers, unnecessary for patch upgrades
    if not patch and controllers_but_master is not None and len(controllers_but_master) > 0:
        enable_extra_controllers = EnableExtraControllersStep(
            admin=admin,
            master_controller=upgrade_plan.master_controller,
            extra_controllers=controllers_but_master
        )
        upgrade_steps.append(enable_extra_controllers)

    # Upgrade log servers that are not also a controller one by one
    if len(upgrade_plan.log_servers) > 0:
        upgrade_log_servers = _upgrade_appliances_step(
            admin=admin,
            appliances=upgrade_plan.log_servers,
            image_url=image_url,
            dev_keyring=dev_keyring,
            starting_status=starting_status,
            throttle=1
        )
        upgrade_steps.append(upgrade_log_servers)
    for appliance_group in upgrade_plan.other_groups:
        # Upgrade gateways that are not also a controller
        # Do the upgrades in two groups, split by site, to minimize downtime.
        if appliance_group:
            upgrade_appliance_group = _upgrade_appliances_step(
                admin=admin,
                appliances=appliance_group,
                image_url=image_url,
                dev_keyring=dev_keyring,
                starting_status=starting_status,
                throttle=throttle
            )
            upgrade_steps.append(upgrade_appliance_group)

    # Upgrade image is not set, that means that we will get it to the controller
    # first, make sure we remove it as well.
    if image_url and image_url.value is None:
        upgrade_steps.append(DeleteUpgradeImageFromControllerStep(
            admin=admin,
            image_url=image_url
        ))
    return upgrade_steps


def filter_appliances_by_target_version(appliances: List[Appliance], starting_status: AppliancesStartingStatus,
                                        target_version: str) -> List[Appliance]:
    # Skip the ones that are already at the target version.
    starting_status.initialize()

    # If the target version is a full version, filter using that, else filter by semantic version.
    if APPLIANCE_VERSION_REGEXP.match(target_version):
        versions = starting_status.versions

        filtered_appliances = [
            appliance for appliance in appliances
            if versions.get(appliance.name) != target_version
        ]
    elif APPLIANCE_SEMVER_REGEXP.match(target_version):
        semantic_versions = starting_status.semantic_versions

        filtered_appliances = [
            appliance for appliance in appliances
            if semantic_versions.get(appliance.name) != target_version
        ]
    else:
        log.warning('Invalid target version %s, will upgrade all appliances', target_version)
        filtered_appliances = appliances

    return filtered_appliances


def filter_offline_appliances(appliances: List[Appliance],
                              starting_status: AppliancesStartingStatus,
                              target_version: Optional[str]) -> List[Appliance]:
    """
    Filter offline appliances from the list of appliances to be upgraded.
    Error out if a controller or a log server are offline as these can cause upgrade issues.
    """
    offline_appliance_names = {
        appliance_name
        for appliance_name, appliance_status in starting_status.status.items()
        if not appliance_status.online
    }
    offline_appliances = {
        appliance for appliance in appliances
        if appliance.name in offline_appliance_names
    }

    # Check for offline controllers
    offline_controllers = {
        appliance for appliance in offline_appliances
        if ApplianceRole.CONTROLLER in appliance.roles
    }
    if len(offline_controllers) > 0:
        require_all_controllers = True
        if target_version is not None:
            # Allow upgrading other appliances if all online controllers are already running the target version,
            # even if some controllers are down.
            online_controller_versions = {
                appliance_status.version
                for appliance_name, appliance_status in starting_status.status.items()
                if (appliance_status.roles and
                    ApplianceRole.CONTROLLER.value in appliance_status.roles and
                    appliance_status.online)
            }
            if online_controller_versions == {target_version}:
                require_all_controllers = False
        if require_all_controllers:
            raise UpgradeError('Cannot start the upgrade since the following controller{}'
                               ' offline: {}'.format(plural(offline_controllers), names(offline_controllers)))

    # Check for offline log servers
    offline_logservers = {
        appliance for appliance in offline_appliances
        if ApplianceRole.LOGSERVER in appliance.roles
    }
    if len(offline_logservers) > 0:
        raise UpgradeError('Cannot start the upgrade since the following log server{}'
                           ' offline: {}'.format(plural(offline_logservers), names(offline_logservers)))

    # Filter out other offline appliances
    if len(offline_appliances) > 0:
        log.warning('%s appliance%s offline: %s', len(offline_appliances),
                    plural(offline_appliances), names(offline_appliances))
        log.warning('%s will be excluded from the upgrade!',
                    'They' if len(offline_appliances) > 1 else 'It')
    return list(filter(lambda a: a.name not in offline_appliance_names,
                       appliances))


def filter_appliances_by_upgrade_status(appliances: List[Appliance],
                                        starting_status: AppliancesStartingStatus,
                                        expected_upgrade_status: Set[UpgradeStatus]) -> List[Appliance]:
    """
    Filter appliances that are not in the expected upgrade status.
    """
    appliances_upgrade_status = starting_status.upgrade_status
    unmatched_appliances = [
        a for a in appliances
        if appliances_upgrade_status[a.appliance_id].status not in expected_upgrade_status
    ]
    if len(unmatched_appliances) == len(appliances):
        status_str = ', '.join(upgrade_status.value for upgrade_status in expected_upgrade_status)
        raise UpgradeError('No appliances are in the expected upgrade status: {}'.format(status_str))
    elif len(unmatched_appliances) > 0:
        appliances_with_status = ', '.join(
            '{} ({})'.format(a.name, appliances_upgrade_status[a.appliance_id].status)
            for a in sorted(unmatched_appliances, key=lambda a: a.name)
        )
        status_str = ', '.join(upgrade_status.value for upgrade_status in expected_upgrade_status)
        log.warning('%d appliance%s not in the expected upgrade status %s: %s',
                    len(unmatched_appliances), plural(unmatched_appliances),
                    status_str, appliances_with_status)
        log.warning('%s will be excluded from the upgrade!',
                    'They' if len(unmatched_appliances) > 1 else 'It')
    return [
        a for a in appliances
        if appliances_upgrade_status[a.appliance_id].status in expected_upgrade_status
    ]


def rollback_upgrade_steps(upgrade_steps: List[UpgradeStep]) -> List[UpgradeStep]:
    steps = (step.rollback() for step in reversed(upgrade_steps))
    return [step for step in steps if step is not None]


def run_steps(upgrade_steps: List[UpgradeStep], rollback_on_failure: bool = True) -> None:
    """
    Run steps one by one, run rollback steps on failure if the flag is set.
    """
    for step_number, upgrade_step in enumerate(upgrade_steps, start=1):
        log.info('%d. %s', step_number, upgrade_step.description)
        try:
            upgrade_step()
        except Exception as e:
            if not rollback_on_failure:
                raise
            rollback_steps = rollback_upgrade_steps(upgrade_steps[:step_number - 1])
            if rollback_steps:
                log.info('')
                log.error('Upgrade step failed with error: %s', e)
                log.error('Rolling back %s applicable step%s',
                          len(rollback_steps), 's' if len(rollback_steps) > 1 else '')
                log.info('')
                try:
                    run_steps(rollback_steps,
                              # Let's avoid infinite recursion...
                              rollback_on_failure=False)
                except CriticalUpgradeError:
                    raise
                except Exception as ce:
                    raise CriticalUpgradeError('Failed to roll back') from ce
            raise
        log.info('')


def show_backup_instructions(master_controller_name: str, yes: bool) -> None:
    log.info('')
    log.info('Please perform a backup or snapshot of %s before continuing!', master_controller_name)
    log.info('Use appgate-backup to perform a backup of the controller.')
    log.info('For more documentation on the backup process, go to:')
    log.info('    %s/backup-script.html', HELP_PREFIX)
    log.info('')
    if not ask_for_confirmation('Have you completed the controller backup or snapshot', yes):
        raise UpgradeCancelled('Cancelled by user pending a controller backup or snapshot!')


def run_prepare_upgrade(admin: ApplianceAdmin, appliances: List[Appliance],
                        master_controller: Appliance, image: ImageSpecifier, *,
                        dev_keyring: bool, target_version: Optional[str] = None, yes: bool = False,
                        throttle: Optional[int] = None) -> None:
    prepared_upgrade_steps = prepare_prepared_upgrade_steps(admin, appliances=appliances,
                                                            master_controller=master_controller, image=image,
                                                            dev_keyring=dev_keyring,
                                                            throttle=throttle, yes=yes)

    # Warn the user to backup if the master controller will be upgraded
    if not yes and master_controller.appliance_id in {appliance.appliance_id for appliance in appliances}:
        show_backup_instructions(master_controller.name, yes)

    # Print the prepared upgrade plan
    log.info('')
    log.info('Prepared upgrade plan:')
    for step_number, prepared_upgrade_step in enumerate(prepared_upgrade_steps, start=1):
        log.info('%d. %s', step_number, prepared_upgrade_step.description)

    if target_version:
        log.info('Appliances will be prepared for upgrade to version %s', target_version)
    log.info('')

    # Ask for user confirmation
    ask_for_upgrade_confirmation(yes=yes)

    # Run the actual upgrade steps!
    run_steps(prepared_upgrade_steps)

    log.notice('Ready for upgrade!')


def run_cancel_upgrade(admin: ApplianceAdmin, appliances: List[Appliance],
                       yes: bool = False, delete_image: Optional[str] = None) -> None:
    # Filter appliances in idle upgrade status
    appliances_upgrade_status = get_appliances_upgrade_status(admin, appliances)
    appliances = [appliance for appliance, upgrade_status in appliances_upgrade_status
                  if upgrade_status.status != UpgradeStatus.IDLE]

    cancel_upgrade_steps = prepare_cancel_upgrade_steps(admin, appliances, delete_image=delete_image)

    if not cancel_upgrade_steps:
        log.notice('No upgrade to cancel')
        return

    # Print the cancel upgrade plan
    log.info('')
    log.info('Cancel upgrade plan:')
    for step_number, prepared_upgrade_step in enumerate(cancel_upgrade_steps, start=1):
        log.info('%d. %s', step_number, prepared_upgrade_step.description)

    log.info('')

    # Ask for user confirmation
    ask_for_upgrade_confirmation(yes=yes)

    # Run the actual upgrade steps!
    run_steps(cancel_upgrade_steps)

    log.notice('Upgrade cancelled!')


def minor_version(semantic_version: str) -> str:
    """
    Return the minor version, for example:
        4.2.1 -> 4.2
        4.2.4 -> 4.2
    """
    return '.'.join(semantic_version.split('.')[:-1])


def is_patch_upgrade(controller_versions: Iterable[str], target_version: str) -> bool:
    """
    An upgrade is a patch if only the patch version changes between the target version and
    all the current controller versions.
    """
    minor_versions = set(map(minor_version, controller_versions))
    if len(minor_versions) != 1:
        return False
    return next(iter(minor_versions)) == minor_version(target_version)


def run_upgrade(admin: ApplianceAdmin, appliances: List[Appliance],
                master_controller: Appliance, all_controllers: List[Appliance],
                starting_status: AppliancesStartingStatus, image: Optional[ImageSpecifier], site_steps: int,
                dev_keyring: bool,
                target_version: Optional[str] = None, yes: bool = False, throttle: Optional[int] = None) -> None:
    """
    Do the upgrade!

    Uses the controller rest API to get the list of appliances, check appliance status etc...
    Schedules the upgrades in the right order: controllers first, then log servers,
    gateways split by site in two groups, and finally appliances with no roles.

    In case this is a multi-controller system, disable all but one controller before upgrading, and re-enable
    them once the controllers have been upgraded.
    """
    upgrade_plan = prepare_upgrade_plan(
        master_controller=master_controller,
        appliances=appliances,
        all_controllers=all_controllers,
        site_steps=site_steps,
    )

    # We are completing a patch upgrade, try to get the target version
    # from the upgrade status.
    if image is None and target_version is None:
        upgrade_statuses = starting_status.upgrade_status.values()
        target_versions = set(filter(None, (
            guess_target_version(upgrade_status.details)
            for upgrade_status in upgrade_statuses
            if upgrade_status.status == UpgradeStatus.READY and upgrade_status.details
        )))
        if len(target_versions) == 1:
            target_version = next(iter(target_versions))
        elif len(target_versions) >= 2:
            log.warning('Found appliances prepared to upgrade to more than one target version: %s',
                        ', '.join(sorted(target_versions)))
            ask_for_upgrade_confirmation(yes=yes)
        else:
            log.warning('Could not guess target version, this upgrade will be performed '
                        'as a major upgrade.')
            ask_for_upgrade_confirmation(yes=yes)

    # Determine if this is a patch upgrade, if so we will not need to disable the
    # extra controllers during the upgrade.
    if target_version is None:
        # In doubt, assume this isn't a patch
        patch = False
    elif any(controller.version < MINIMUM_VERSION_HA_PATCH_CONTROLLER_UPGRADES
             for controller in all_controllers):
        # Older controllers don't support HA patch upgrades.
        patch = False
    else:
        controller_names = {controller.name for controller in upgrade_plan.controllers}
        controllers_versions = {
            version for name, version in starting_status.semantic_versions.items()
            if name in controller_names and version
        }
        patch = is_patch_upgrade(controllers_versions, target_version)

    # Prepare the actual upgrade steps so we can refer to them by index.
    upgrade_steps = prepare_upgrade_steps(admin=admin, upgrade_plan=upgrade_plan,
                                          master_controller=master_controller,
                                          image=image, starting_status=starting_status,
                                          dev_keyring=dev_keyring,
                                          throttle=throttle, patch=patch, yes=yes)

    # Warn the user to backup if the master controller will be upgraded
    if not yes and upgrade_plan.master_controller is not None:
        show_backup_instructions(upgrade_plan.master_controller.name, yes)

    # Print the upgrade plan
    log.info('')
    log.info('Upgrade plan:')
    for step_number, upgrade_step in enumerate(upgrade_steps, start=1):
        log.info('%d. %s', step_number, upgrade_step.description)

    if target_version:
        log.info('Appliances will be upgraded to version %s%s',
                 target_version, ' (patch upgrade)' if patch else '')
    log.info('')

    # Ask for user confirmation
    ask_for_upgrade_confirmation(yes=yes)

    # Run the actual upgrade steps!
    run_steps(upgrade_steps)

    log.notice('Upgrade done!')
    upgrade_script_download_link = admin.url + '/appliance/upgrade-script'
    log.info('Upgraded using appgate-upgrade %s, the latest version is available on your controller at %s',
             VERSION, upgrade_script_download_link)

    # Finally, display version information:
    appliances_status = get_appliances_status(admin)
    versions = {s.version for s in appliances_status.values()}
    if len(versions) == 1:
        version = next(iter(versions))
        log.notice('All appliances are now running version %s', version)
    else:
        log.warning('The upgrade completed but not all appliances are running the same version!')
        for name, appliance_status in appliances_status.items():
            appliance_version = appliance_status.version or 'unknown'
            log.info('Appliance %s now running version %s', name, appliance_version)


def show_upgrade_status(admin: ApplianceAdmin, appliances: List[Appliance]) -> None:
    log.info('')
    log.info('Appliance%s upgrade status:', 's' if len(appliances) > 1 else '')
    appliances_upgrade_status = get_appliances_upgrade_status(admin, appliances)
    for appliance, appliance_status in sorted(appliances_upgrade_status, key=lambda a: a[0].name):
        line = '{}: {} ({})'.format(appliance.name, appliance_status.status.value,
                                    appliance_status.details or 'no details')
        if appliance_status.status in {UpgradeStatus.READY, UpgradeStatus.SUCCESS}:
            log.notice(line)
        elif appliance_status.status == UpgradeStatus.IDLE:
            log.info(line)
        elif appliance_status.status == UpgradeStatus.FAILED:
            log.error(line)
        else:
            log.warning(line)


class UpgradeCommand(enum.Enum):
    INSTALL_UPGRADE = 'install'
    PREPARE_UPGRADE = 'prepare'
    COMPLETE_UPGRADE = 'complete'
    CANCEL_UPGRADE = 'cancel'
    UPGRADE_STATUS = 'status'


def add_common_positional_args(parser) -> None:
    parser.add_argument('controller', metavar='CONTROLLER',
                        help=os.linesep.join([
                            'Appgate SDP controller peer or admin hostname or ip',
                            'The server should be specified in the following form:',
                            '    [https://]hostname[:port]',
                            'The hostname should point to a specific controller.',
                            'If the port is specified here it takes precedence over the port argument.',
                            'The port should be the Appgate SDP admin API port.',
                            'If the admin interface is configured, use admin interface port.',
                            'Else, use the peer interface port.'
                        ]))


def add_common_optional_args(parser) -> None:
    common_args = parser.add_argument_group('common arguments')
    common_args.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
                             help='Show this help message and exit')
    common_args.add_argument('--username', default='admin',
                             help=os.linesep.join([
                                 'Username of the upgrade admin user',
                                 'The password can be set via the APPGATE_PASSWORD environment variable, '
                                 'it will be prompted if not set.'
                             ]))
    common_args.add_argument('--provider', default='local', help='Identity provider for the upgrade admin user')
    common_args.add_argument('--cacert', default=None,
                             help='Path to the controller\'s CA cert file in PEM or DER format')
    common_args.add_argument(
        '--site-steps',
        help='Specify the desired amount of steps to upgrade the appliances in a site [default=2]',
        type=int,
        default=2,
    ),

    # Occasional usage or debugging purposes
    advanced_args = parser.add_argument_group('advanced arguments')
    advanced_args.add_argument('--no-verify', action='store_true', default=False,
                               help='Do NOT verify the controller\'s certificate')
    advanced_args.add_argument('--dev-keyring', action='store_true', default=False,
                               help='Use the development keyring to verify the upgrade image')
    advanced_args.add_argument('-y', '--yes', default=False, action='store_true',
                               help='Batch mode, don\'t ask for confirmation')
    advanced_args.add_argument('--filter', default=None, type=str, action='append',
                               help=os.linesep.join([
                                   'Include only appliance matching the given key=value filter, '
                                   'e.g. site=06590afe-3eae-4437-aff6-523c06045103',
                                   'The value of an filter supports wildcards using *, '
                                   'e.g. --filter \'name=example*\'',
                                   'Valid filter attributes are: {}'.format(
                                       ', '.join(sorted(APPLIANCE_FILTER_ATTRIBUTES)))
                               ]))
    advanced_args.add_argument('--exclude', default=None, type=str, action='append',
                               help=os.linesep.join([
                                   'Exclude appliances matching the given key=value filter, e.g. role=controller',
                                   'The value of an exclude supports wildcards using *, '
                                   'e.g. --exclude \'name=example*\'',
                                   'Valid exclude attributes are: {}'.format(
                                       ', '.join(sorted(APPLIANCE_FILTER_ATTRIBUTES)))
                               ]))
    advanced_args.add_argument('--target-version',
                               help=os.linesep.join([
                                   'Specify the target Appgate SDP version in X.Y.Z format (e.g. 3.2.1).',
                                   'Alternatively the X.Y.Z-buildnumber-branch format (e.g. 3.2.1-6220-release) '
                                   'can also be used.',
                                   'The target version will be guessed from the image URL if it is unspecified'
                               ]),
                               type=str, default=None)
    advanced_args.add_argument('--ignore-target-version',
                               help=os.linesep.join([
                                   'Do not try to guess the target version from the image URL.',
                                   'This takes precedence over --target-version if both are set'
                               ]),
                               action='store_true')
    advanced_args.add_argument('--peer-api-version', default=None, type=int, required=False,
                               help='Controller Peer API version, it will be automatically discovered if not set')
    advanced_args.add_argument('--actual-hostname',
                               help=os.linesep.join([
                                   'Specifies the actual hostname used in the admin (or peer) '
                                   'interface of this Appgate SDP controller.',
                                   'ONLY use this option when connecting using an ip or hostname '
                                   'that differs from the actual hostname.'
                                   'This could be the case when the system has been configured for NAT.'
                               ]),
                               type=str, default=None)
    advanced_args.add_argument('--no-color', help='Do not use colors', action='store_true')
    advanced_args.add_argument('--device-id', default=None,
                               help='Device id of the local machine, it will be generated at every run if not set')
    advanced_args.add_argument('-l', '--log-level', help='Log level',
                               choices=list(sorted(LOG_LEVEL_MAPPING.keys())))


def check_prepare_timeout(value: str) -> float:
    int_value = float(value)
    if int_value < MINIMUM_PREPARE_TIMEOUT:
        raise argparse.ArgumentTypeError('Invalid prepare timeout, should be at least %.1f' % MINIMUM_PREPARE_TIMEOUT)
    return int_value


def add_image_arguments(parser) -> None:
    image_args = parser.add_argument_group('image arguments')
    image_args.add_argument('image', metavar='IMAGE', help='Upgrade image file or URL')
    image_args.add_argument('--host-on-controller', action='store_true', default=False,
                            help='Host upgrade image file on controller')
    image_args.add_argument('--throttle', metavar='THROTTLE', type=int, default=DEFAULT_THROTTLE,
                            help='Throttle upgrade image downloads, THROTTLE appliances at a time, '
                                 'use 0 to disable throttling')
    image_args.add_argument('--prepare-timeout', type=check_prepare_timeout, default=DEFAULT_PREPARE_TIMEOUT,
                            help='Time in seconds after which the upgrade script will timeout when preparing '
                                 'an appliance for upgrade')


def check_complete_timeout(value: str) -> float:
    int_value = float(value)
    if int_value < MINIMUM_COMPLETE_TIMEOUT:
        raise argparse.ArgumentTypeError('Invalid complete timeout, should be at least %.1f' % MINIMUM_COMPLETE_TIMEOUT)
    return int_value


def add_installation_arguments(parser) -> None:
    complete_args = parser.add_argument_group('installation arguments')
    complete_args.add_argument('--complete-timeout', type=check_complete_timeout, default=DEFAULT_COMPLETE_TIMEOUT,
                               help='Time in seconds after which the upgrade script will timeout when completing the '
                                    'upgrade installation on an appliance')


def args_parser(command_name: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=os.linesep.join([
            'Appgate SDP upgrade script.',
            os.linesep,
            '© 2020 Cyxtera Cybersecurity, Inc. d/b/a Appgate',
            'All rights reserved. Appgate is a trademark of Cyxtera Cybersecurity, Inc. d/b/a Appgate',
            os.linesep,
            'https://www.appgate.com',
            os.linesep,
            'For more documentation on the upgrade process, go to:',
            '    {}/upgrading-appliances.html?anchor=collective-upgrade'.format(HELP_PREFIX)
        ]),
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=VERSION)

    subparsers = parser.add_subparsers(help='Run the given upgrade command',
                                       dest='command', title='upgrade command')
    subparsers.required = True

    install_description = os.linesep.join([
        'Perform a full upgrade.',
    ])
    install_parser = subparsers.add_parser('install', description=install_description,
                                           add_help=False, formatter_class=argparse.RawTextHelpFormatter)
    install_parser.set_defaults(command=UpgradeCommand.INSTALL_UPGRADE)
    add_common_positional_args(install_parser)
    add_image_arguments(install_parser)
    add_installation_arguments(install_parser)
    add_common_optional_args(install_parser)

    prepare_description = os.linesep.join([
        'Prepare an upgrade but do NOT install it.',
        os.linesep,
        'This means the upgrade file will be downloaded/uploaded to all the appliances,',
        'the signature verified as well as any other preconditions applicable at this point.'
    ])
    prepare_parser = subparsers.add_parser('prepare', description=prepare_description,
                                           add_help=False, formatter_class=argparse.RawTextHelpFormatter)
    prepare_parser.set_defaults(command=UpgradeCommand.PREPARE_UPGRADE)
    add_common_positional_args(prepare_parser)
    add_image_arguments(prepare_parser)
    add_common_optional_args(prepare_parser)

    complete_description = os.linesep.join([
        'Complete a prepared upgrade.',
        os.linesep,
        'Install a prepared upgrade on the secondary partition ',
        'and perform a reboot to make the second partition the primary.'
    ])
    complete_parser = subparsers.add_parser('complete', description=complete_description,
                                            add_help=False, formatter_class=argparse.RawTextHelpFormatter)
    complete_parser.set_defaults(command=UpgradeCommand.COMPLETE_UPGRADE)
    add_common_positional_args(complete_parser)
    add_installation_arguments(complete_parser)
    add_common_optional_args(complete_parser)

    cancel_parser = subparsers.add_parser('cancel', description='Cancel a prepared upgrade',
                                          add_help=False, formatter_class=argparse.RawTextHelpFormatter)
    cancel_parser.add_argument('--delete-image-from-controller', type=str, default=None,
                               help='Delete upgrade image file from controller')
    cancel_parser.set_defaults(command=UpgradeCommand.CANCEL_UPGRADE)
    add_common_positional_args(cancel_parser)
    add_common_optional_args(cancel_parser)

    status_parser = subparsers.add_parser('status', description='Display the upgrade status',
                                          add_help=False, formatter_class=argparse.RawDescriptionHelpFormatter)
    status_parser.set_defaults(command=UpgradeCommand.UPGRADE_STATUS)
    add_common_positional_args(status_parser)
    add_common_optional_args(status_parser)
    return parser


def parse_args(argv: List[str]):
    command_name = Path(argv[0]).name
    parser = args_parser(command_name)
    return parser.parse_args(args=argv[1:])


def der_to_pem_file(path: str) -> Optional[tempfile.NamedTemporaryFile]:  # type: ignore
    """
    Convert a DER certificate to PEM.

    Return None if it was already a PEM file.
    Return a handle to the PEM temporary file if it was a DER certificate.
    """
    with open(path, 'rb') as f:
        content = f.read()

    # If the file doesn't begin with a PEM header, treat it as DER
    # and convert to PEM. Use the resulting temporary PEM file.
    try:
        if content[0:len(ssl.PEM_HEADER)].decode() == ssl.PEM_HEADER:
            return None
    except UnicodeDecodeError:
        # Binary content, assume DER
        pass

    pem_content = ssl.DER_cert_to_PEM_cert(content)
    pem_file_name = Path(path).with_suffix('.pem').name
    pem_file = tempfile.NamedTemporaryFile(suffix=pem_file_name, mode='w+t')
    pem_file.write(pem_content)
    pem_file.seek(0)

    # Return the temporary file handle itself so it doesn't
    # get garbage collected.
    return pem_file


def check_throttle(throttle: Optional[int], num_appliance: int) -> None:
    """
    Emit a warning if the effective throttle is higher than the default.
    """
    if throttle is None:
        effective_throttle = num_appliance
    else:
        effective_throttle = min(throttle, num_appliance)

    if effective_throttle > DEFAULT_THROTTLE:
        log.warning('The appliance download throttle is set to %d, meaning up to %d appliances will download '
                    'the upgrade image at the same time, this will likely result in download failures',
                    throttle if throttle is not None else 0, effective_throttle)


def run_upgrade_command(args) -> None:
    upgrade_command = UpgradeCommand(args.command)
    dev_keyring = args.dev_keyring

    if os.path.exists('/mnt/state/config'):
        raise Exception('This should not be executed on an appliance')

    if not args.no_color:
        pick_color_function()

    try:
        hostname, port = parse_server(server=args.controller)
    except ValueError as e:
        raise UpgradeError(str(e))

    controller_url = 'https://{}:{}'.format(hostname, port)

    username = args.username
    provider = args.provider
    device_id = args.device_id or default_device_id()
    if args.no_verify:
        verify = False
    elif args.cacert:
        verify = args.cacert
    else:
        verify = True
    peer_api_version = args.peer_api_version
    if args.log_level:
        set_log_level(args.log_level)

    target_version = args.target_version  # type: Optional[str]
    ignore_target_version = args.ignore_target_version  # type: bool
    yes = args.yes  # type: bool
    image_arg = getattr(args, 'image', None)  # type: Optional[str]

    throttle_arg = getattr(args, 'throttle', 0)  # type: int
    # Any number <= 0 disables throttling
    throttle = throttle_arg if throttle_arg > 0 else None

    site_steps = args.site_steps

    if hasattr(args, 'prepare_timeout'):
        TIMEOUTS.prepare = args.prepare_timeout

    if hasattr(args, 'complete_timeout'):
        TIMEOUTS.complete = args.complete_timeout

    image = None  # type: Optional[ImageSpecifier]
    if image_arg is not None:
        if os.path.exists(image_arg):
            # Image is a file
            image = ImageFilePath(Path(image_arg))

            # Checking that it is a zip file
            with open(image_arg, 'rb') as f:
                header = f.read(4)
            if header != b'PK\x03\x04':
                raise Exception('The image file must be a zip file.')
        else:
            # Image is a URL
            url = urllib.parse.urlparse(image_arg)
            if not all([url.scheme, url.netloc, url.path]):
                raise UpgradeError('Not a valid upgrade image url: {}'.format(image_arg))
            image = ImageURL(image_arg, host_on_controller=args.host_on_controller)

    validate_hostname_uniqueness(hostname, port)

    actual_hostname = hostname
    if args.actual_hostname is not None:
        # The user can select a different hostname than the one used to establish the connection, if connecting
        # e.g. via an internal ip.
        actual_hostname = args.actual_hostname

    if isinstance(verify, str):
        if not os.path.exists(verify):
            raise UpgradeError('Controller certificate file not found: {}'.format(verify))

    # Convert eventual DER cert to PEM.
    if verify and os.path.isfile(verify):  # type: ignore
        pem_file = der_to_pem_file(verify)  # type: ignore
        if pem_file is not None:
            verify = pem_file.name

    # Setup admin object to handle all communications with the controller.
    admin = ApplianceAdmin(controller_url, verify, peer_api_version=peer_api_version,
                           # Do 10 tries for each admin HTTP request by default
                           tries=10)
    try:
        admin.authenticate(username, provider, device_id, os.getenv('APPGATE_PASSWORD'))
    except SSLError as e:
        raise UpgradeError('Certificate verification failed, use the --cacert option to provide the controller\'s '
                           'CA certificate.\n\tSSLError: %s' % e)
    admin.authorize()

    if admin.peer_api_version < MINIMUM_PEER_API_VERSION:
        log.warning('The Appgate appliance upgrade script is not as reliable for peer API '
                    ' versions less than %s, current API version is %s,',
                    MINIMUM_PEER_API_VERSION, admin.peer_api_version)

    # Get the list of all active appliances
    appliances = admin.appliances()

    # Verify that the hostname matches a controller admin hostname
    master_controller = find_master_controller(actual_hostname, appliances)

    # Prepare upgrade plan and ask the user for confirmation
    starting_status = AppliancesStartingStatus(admin=admin, appliances=appliances)

    # Gather the starting status so we know what appliances are offline and what version
    # the online appliances are running before starting the upgrade.
    starting_status.initialize()

    # Figure out the target version
    if ignore_target_version:
        target_version = None
    elif target_version is None and image_arg is not None:
        target_version = guess_target_version(image_arg)

    # Keep a list of all controllers, even before filtering.
    # We need to be able to disable the other controllers even when not upgrading them.
    all_controllers = [appliance for appliance in appliances
                       if ApplianceRole.CONTROLLER in appliance.roles]
    # Filter offline appliances
    appliances = filter_offline_appliances(appliances=appliances, starting_status=starting_status,
                                           target_version=target_version)
    if not appliances:
        # This should never happen
        raise UpgradeError('All appliances are offline!')

    # Filter appliances by target version
    if target_version is not None:
        appliances = filter_appliances_by_target_version(appliances=appliances,
                                                         starting_status=starting_status,
                                                         target_version=target_version)
        if not appliances:
            log.notice('All appliances are already running target version %s', target_version)
            return

    # Filter appliances based on user provided filters
    user_filter = build_appliance_filter(filters=args.filter, excludes=args.exclude)
    appliances = list(user_filter(appliances))

    # Filter the appliances in the starting status too.
    starting_status = starting_status.filter_appliances(appliances)

    if upgrade_command in {UpgradeCommand.INSTALL_UPGRADE, UpgradeCommand.PREPARE_UPGRADE}:
        check_throttle(throttle=throttle, num_appliance=len(appliances))

    # Run the given upgrade command
    if upgrade_command == UpgradeCommand.INSTALL_UPGRADE:
        assert image is not None
        run_upgrade(admin, appliances=appliances, master_controller=master_controller,
                    starting_status=starting_status, yes=yes,
                    target_version=target_version, image=image,
                    site_steps=site_steps,
                    throttle=throttle if throttle else None,
                    dev_keyring=dev_keyring,
                    all_controllers=all_controllers)
    elif upgrade_command == UpgradeCommand.PREPARE_UPGRADE:
        assert image is not None
        run_prepare_upgrade(admin, appliances=appliances, master_controller=master_controller,
                            dev_keyring=dev_keyring,
                            yes=yes, target_version=target_version,
                            image=image, throttle=throttle if throttle else None)
    elif upgrade_command == UpgradeCommand.COMPLETE_UPGRADE:
        assert image is None
        # Filter appliances bad on the upgrade status
        appliances = filter_appliances_by_upgrade_status(appliances=appliances,
                                                         starting_status=starting_status,
                                                         expected_upgrade_status={UpgradeStatus.READY,
                                                                                  UpgradeStatus.SUCCESS})
        run_upgrade(admin, appliances=appliances, master_controller=master_controller,
                    starting_status=starting_status, yes=yes, target_version=target_version,
                    image=None,
                    dev_keyring=dev_keyring,
                    site_steps=site_steps,
                    all_controllers=all_controllers)
    elif upgrade_command == UpgradeCommand.CANCEL_UPGRADE:
        delete_image_from_controller = args.delete_image_from_controller  # type: Optional[str]
        run_cancel_upgrade(admin=admin, appliances=appliances, yes=yes, delete_image=delete_image_from_controller)
    elif upgrade_command == UpgradeCommand.UPGRADE_STATUS:
        show_upgrade_status(admin, appliances)
    else:
        raise UpgradeError('Invalid upgrade command: {}'.format(upgrade_command.value))


def main():
    # Special case to allow --version flag
    if len(sys.argv) == 2 and sys.argv[1] == '--version':
        print(VERSION)
        sys.exit(0)

    args = parse_args(sys.argv)
    try:
        run_upgrade_command(args)
    except KeyboardInterrupt:
        log.warning('\nInterrupted by user')
        sys.exit(2)
    except UpgradeCancelled as e:
        log.error(str(e))
        sys.exit(2)
    except CriticalUpgradeError as e:
        log.exception('Upgrade failed and automated rollback steps could not be completed: %s', e)
        sys.exit(1)
    except (UpgradeError, APIVersionError, AdminError) as e:
        log.error('Upgrade failed: %s', e)
        sys.exit(1)
    except Exception as e:
        log.exception('Unexpected error during upgrade: %s', e)
        sys.exit(1)


if __name__ == '__main__':
    main()
