import enum
import getpass
import http.client
import time
import uuid
from json import loads as json_loads
from json import dumps as json_dumps
import os
import socket
import ssl
from typing import Optional, Dict, Union, Any, Tuple, Iterable, Callable
import urllib.error
import urllib.parse
import urllib.request

from .logger import log

__all__ = [
    'PEER_API_VERSION',
    'MINIMUM_VERSION_FILE_REPOSITORY_API',
    'AdminError',
    'APIVersionError',
    'Response',
    'URLError',
    'HTTPError',
    'SSLError',
    'URLConnectionError',
    'URLConnectionRefusedError',
    'ask_for_password',
    'OTPInit',
    'ChallengeOTPInit',
    'PushOTPInit',
    'FileStatus',
    'DetailedFileStatus',
    'ProgressCallback',
    'Admin',
    'multipart',
]


# Most recent peer API version supported
PEER_API_VERSION = 14


class AdminError(Exception):
    pass


class APIVersionError(AdminError):
    def __init__(self, actual: int, minimum: int) -> None:
        super().__init__('API not supported for peer API version {}, minimum {} required'.format(actual, minimum))
        self.actual = actual
        self.minimum = minimum


def _read_response(r: http.client.HTTPResponse) -> str:
    return r.read().decode('utf-8')


class Response:
    def __init__(self, status_code: int, content: str) -> None:
        self.status_code = status_code
        self.content = content

    @staticmethod
    def from_urllib_error(e: urllib.error.HTTPError) -> 'Response':
        return Response(status_code=e.code, content=e.fp.read().decode('utf-8'))  # type: ignore

    @staticmethod
    def from_http_response(r: http.client.HTTPResponse) -> 'Response':
        return Response(status_code=r.code, content=_read_response(r))  # type: ignore

    def json(self) -> dict:
        return json_loads(self.content)


class URLError(AdminError):
    pass


class HTTPError(URLError):
    """
    Mostly requests compatible error class to wrap urllib errors.
    To help porting code from requests to urllib.
    """

    def __init__(self, response: Response) -> None:
        super().__init__('HTTP Error {}: {}'.format(response.status_code, response.content))
        self.response = response

    @staticmethod
    def from_urllib_error(e: urllib.error.HTTPError) -> 'HTTPError':
        return HTTPError(Response.from_urllib_error(e))

    @staticmethod
    def from_http_response(r: http.client.HTTPResponse) -> 'HTTPError':
        return HTTPError(Response.from_http_response(r))


class SSLError(URLError):
    pass


class URLConnectionError(URLError):
    pass


class URLConnectionRefusedError(URLConnectionError):
    def __str__(self) -> str:
        return 'Connection refused'


class URLConnectionTimeoutError(URLConnectionError):
    def __str__(self) -> str:
        return 'Connection timed out'


def ask_for_password(username: str, provider: str) -> str:
    return getpass.getpass('Enter password for username {} '
                           'and provider name {}:'.format(username, provider))


class OTPInit:
    pass


class ChallengeOTPInit(OTPInit):
    def __init__(self, state: Optional[str] = None) -> None:
        super().__init__()
        self.state = state


class PushOTPInit(OTPInit):
    def __init__(self, send_password: bool, timeout: int) -> None:
        super().__init__()
        self.send_password = send_password
        self.timeout = timeout


class FileStatus(enum.Enum):
    IN_PROGRESS = 'InProgress'
    READY = 'Ready'
    FAILED = 'Failed'


class DetailedFileStatus:
    def __init__(self,
                 status: FileStatus,
                 checksum: Optional[str] = None,
                 failure_reason: Optional[str] = None) -> None:
        self.status = status
        self.checksum = checksum
        self.failure_reason = failure_reason

    def __repr__(self) -> str:
        return '{}(status={!r}, checksum={!r}, failure_reason={!r})'.format(
            self.__class__.__name__, self.status, self.checksum, self.failure_reason)


def multipart_body(files: Dict[str, Tuple[str, str, str]], boundary: str,
                   *, iter_body: bool = True) -> Iterable[bytes]:
    # Start boundary
    yield b'--%b\r\n' % boundary.encode('utf-8')

    # Files
    for stream_name, file_data in files.items():
        name, file, file_content_type = file_data
        # File headers.
        file_headers = [
            ('Content-Type', file_content_type),
            ('Content-Disposition', 'form-data; name="{}"; filename="{}"'.format(stream_name, name))
        ]
        for key, value in file_headers:
            yield b'%b: %b\r\n' % (key.encode('utf-8'), value.encode('utf-8'))
        yield b'\r\n'

        # File body
        if iter_body:
            with open(file, 'rb') as f:
                while True:
                    data = f.read(1048576)
                    if not data:
                        break
                    yield data
        yield b'\r\n'

    # End boundary
    yield b'--%b--\r\n' % boundary.encode('utf-8')


ProgressCallback = Callable[[int, int], None]


def multipart_body_with_progress(files: Dict[str, Tuple[str, str, str]], boundary: str,
                                 content_length: int, progress_callback: ProgressCallback) -> Iterable[bytes]:
    for data in multipart_body(files, boundary):
        progress_callback(len(data), content_length)
        yield data


def multipart(files: Dict[str, Tuple[str, str, str]],
              boundary: Optional[str] = None,
              *, progress_callback: Optional[ProgressCallback] = None) -> Tuple[Dict[str, str], Iterable[bytes]]:
    """
    urllib doesn't support multipart upload, we can trick it into doing them anyways by generating the multipart
    body and supplying the content length. This only works with files of known length.
    """
    if boundary is None:
        boundary = uuid.uuid4().hex

    # To calculate the content length, generate the body without the files content and count it's length.
    # Add to it the size of all files.
    files_size = sum(os.path.getsize(filename) for _, filename, _ in files.values())
    content_length = sum(map(len, multipart_body(files, boundary=boundary, iter_body=False))) + files_size

    headers = {
        'Content-Type': 'multipart/form-data; boundary={}'.format(boundary),
        'Content-Length': str(content_length)
    }
    if progress_callback is None:
        return headers, multipart_body(files, boundary)
    return headers, multipart_body_with_progress(files, boundary, content_length, progress_callback)


URLData = Union[bytes, Iterable[bytes]]


# The file repository API was added in peer API version 7
MINIMUM_VERSION_FILE_REPOSITORY_API = 7


# Always set default timeout, urllib.request can hang *forever* in some situations
# when timeout is left set to socket._GLOBAL_DEFAULT_TIMEOUT.
DEFAULT_SOCKET_TIMEOUT = 60.0


class Admin:
    """
    The Admin class is used to handle all interactions with the controller REST API necessary
    for upgrades.

    :param tries: Default number of tries unless otherwise specified in the request call.
    The default value is 1 which effectively disables retries.
    """

    def __init__(self, url: str, verify: Union[bool, str], peer_api_version: Optional[int] = None,
                 *, tries: int = 1) -> None:
        self.url = url
        self.verify = verify

        # If the peer api version is None it will be discovered automatically.
        self._peer_api_version = peer_api_version
        self._auth_token = None  # type: Optional[str]

        # Saved credentials for re-authentication
        self._username = None  # type: Optional[str]
        self._provider = None  # type: Optional[str]
        self._device_id = None  # type: Optional[str]
        self._password = None  # type: Optional[str]

        # Default number of tries of requests
        self._tries = tries

    @property
    def peer_api_version(self) -> int:
        if self._peer_api_version is None:
            raise Exception('No peer API version set!')
        return self._peer_api_version

    @property
    def accept_header(self) -> str:
        header_format = 'application/vnd.cryptzone.peer-v{}+json' if self.peer_api_version < 7 else 'application/vnd.appgate.peer-v{}+json'
        return header_format.format(self.peer_api_version)

    def _request(self, method: str, path: str, *, json: dict = None, headers: Dict[str, str] = None,
                 files: Optional[Dict[str, Tuple[str, str, str]]] = None,
                 timeout: float = DEFAULT_SOCKET_TIMEOUT, outfile: Optional[str] = None,
                 progress_callback: Optional[ProgressCallback] = None) -> Dict[str, Any]:
        """
        This method provides a requests-like abstraction on top of the stdlib urllib.

        If outfile is specified, the response is written in the file rather than returned, and an
        empty dictionary is returned.
        """
        headers = headers or {}
        if 'Accept' not in headers:
            headers['Accept'] = self.accept_header
        if 'Authorization' not in headers and self._auth_token:
            headers['Authorization'] = 'Bearer {}'.format(self._auth_token)

        data = None  # type: Optional[URLData]
        if json is not None:
            data = json_dumps(json).encode('utf-8')
            headers['Content-Type'] = 'application/json;'
            headers['Content-Length'] = str(len(data))
        elif files is not None:
            extra_headers, data = multipart(files, progress_callback=progress_callback)
            headers.update(extra_headers)

        req = urllib.request.Request('{}{}'.format(self.url, path), method=method, data=data)  # type: ignore
        for key, value in headers.items():
            req.add_header(key, value)

        ssl_context = None  # type: Any
        cafile = None  # type: Any
        if isinstance(self.verify, str):
            cafile = self.verify
        elif not self.verify:
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

        try:
            r = urllib.request.urlopen(req, cafile=cafile, context=ssl_context, timeout=timeout)
        except urllib.error.HTTPError as e:
            if e.args and isinstance(e.args[0], ssl.SSLError):
                raise
            raise HTTPError.from_urllib_error(e) from e
        except urllib.error.URLError as e:
            if e.args:
                cause = e.args[0]
                if isinstance(cause, ssl.SSLError):
                    raise SSLError(e.args) from e
                elif isinstance(cause, ConnectionRefusedError):
                    raise URLConnectionRefusedError(e.args) from e
                elif isinstance(cause, ConnectionError):
                    raise URLConnectionError(e.args) from e
                elif isinstance(cause, TimeoutError):
                    raise URLConnectionTimeoutError(e.args) from e

            # Consider all other errors as an url error.
            raise URLError(e.args) from e

        if r.status not in {200, 202, 204}:
            raise HTTPError.from_http_response(r)
        elif r.status in {200, 202} and outfile:
            with open(outfile, 'wb') as f:
                while True:
                    buffer = r.read(1024)
                    if len(buffer) == 0:
                        break
                    f.write(buffer)
            return {}
        elif r.status in {200, 202}:
            content = r.read().decode('utf-8')
            if content:
                return json_loads(content)
            else:
                return {}
        else:
            return {}

    def _authenticate(self, username: str, password: str, provider: str, device_id: str) -> str:
        """
        Authenticates to the controller.
        Returns the auth token as a base64 string on success.
        """
        credentials = {
            'username': username,
            'password': password,
            'providerName': provider,
            'deviceId': device_id
        }
        auth_token = self._request('POST', '/admin/authentication', json=credentials)['token']
        return auth_token

    def _discover_peer_api_version(self) -> int:
        """
        Discover the peer API version by doing a request with a wrong request header
        and extracting the peer API version from the error response.
        """
        try:
            self._request('POST', '/admin/authentication', headers={'Accept': 'application/json'})
        except HTTPError as e:
            if e.response.status_code == 406:
                # Pick the highest peer API version supported by both the controller and the admin client.
                max_peer_api_version = e.response.json()['maxSupportedVersion']
                min_peer_api_version = e.response.json()['minSupportedVersion']
                peer_api_version = min(PEER_API_VERSION, max_peer_api_version)
                if peer_api_version < min_peer_api_version:
                    raise AdminError('Could not find a compatible peer API version!')
                return peer_api_version
            else:
                raise
        raise AdminError('Failed to discover peer API version!')

    def authenticate(self, username: str, provider: str, device_id: str,
                     password: Optional[str] = None, num_attempts: int = 3) -> None:
        if self._peer_api_version is None:
            log.info('Discovering peer API version for %s...', self.url)
            try:
                self._peer_api_version = self._discover_peer_api_version()
            except HTTPError as e:
                try:
                    error = 'Peer API version discovery failed: {}'.format(e.response.json()['message'])
                except (ValueError, KeyError):
                    error = 'Peer API version discovery failed with HTTP {}: {}'.format(
                        e.response.status_code, e.response.content)
                raise AdminError(error) from e
            log.info('Peer API version is %s', self._peer_api_version)

        if password is None:
            password = ask_for_password(username, provider)

        for _ in range(num_attempts):
            try:
                log.info('Authenticating to %s...', self.url)
                self._auth_token = self._authenticate(username, password, provider, device_id)

                # All went well, save credentials for a possible later re-authentication
                self._username = username
                self._provider = provider
                self._device_id = device_id
                self._password = password

                log.notice('Authentication succeeded')
                return
            except HTTPError as e:
                if e.response.status_code == 401:
                    log.warning('Invalid username or password!')
                    password = ask_for_password(username, provider)
                    continue
                else:
                    try:
                        error = 'Authentication failed: {}'.format(e.response.json()['message'])
                    except (ValueError, KeyError):
                        error = 'Authentication failed with HTTP {}: {}'.format(
                            e.response.status_code, e.response.content)
                    raise AdminError(error) from e
        raise AdminError('Authentication failed!')

    def _initialize_otp(self) -> OTPInit:
        """
        Initialize OTP and guide the user through the appropriate steps.
        Optionally returns a state to pass to the OTP request.
        """
        otp_init = self._request('POST', '/admin/authentication/otp/initialize',
                                 json={'userPassword': self._password})
        otp_type = otp_init.get('type')
        if otp_type == 'Secret':
            log.warning('One-time password initialization is required!')
            log.info('To initialize the timed based OTP enter the following secret: %s', otp_init['secret'])
            return ChallengeOTPInit()
        elif otp_type == 'RadiusChallenge':
            return ChallengeOTPInit(state=otp_init['state'])
        elif otp_type == 'AlreadySeeded':
            # Time based OTP already seeded, carry on...
            return ChallengeOTPInit()
        elif otp_type == 'Push':
            return PushOTPInit(send_password=otp_init['sendPassword'],
                               timeout=otp_init['timeout'])
        else:
            raise AdminError('OTP type {} not supported!'.format(otp_type))

    def _authenticate_radius_challenge_otp(self, otp_init: ChallengeOTPInit,
                                           otp: Optional[str] = None) -> str:
        if otp is None:
            otp = getpass.getpass('Please enter your one-time password:')
        body = {
            'otp': otp
        }
        if otp_init.state is not None:
            body['state'] = otp_init.state
        auth_token = self._request('POST', '/admin/authentication/otp', json=body)['token']
        return auth_token

    def _authenticate_push_otp(self, otp_init: PushOTPInit) -> str:
        body = {}
        if otp_init.send_password:
            body['otp'] = self._password
        else:
            body['otp'] = 'push'

        # Give ourselves 10 seconds more than the server side timeout
        timeout = otp_init.timeout + 10
        print('Open the two-factor authentication app on your device and approve your login.')
        auth_token = self._request('POST', '/admin/authentication/otp', json=body, timeout=timeout)['token']
        return auth_token

    def authenticate_otp(self, num_attempts: int = 3) -> None:
        otp_init = self._initialize_otp()

        # Give the user a few attempts to get the OTP right...
        for _ in range(num_attempts):
            try:
                if isinstance(otp_init, ChallengeOTPInit):
                    self._auth_token = self._authenticate_radius_challenge_otp(otp_init)
                elif isinstance(otp_init, PushOTPInit):
                    self._auth_token = self._authenticate_push_otp(otp_init)
                else:
                    raise AdminError('Invalid otp init of type {}: {}'.format(type(otp_init), otp_init))
                break
            except HTTPError as e:
                if e.response.status_code == 401:
                    error = e.response.json()
                    message = error.get('message', 'Invalid OTP!')
                    log.warning('%s', message)
                else:
                    raise
        else:
            raise AdminError('One-time password authentication failed!')

    def _authorize(self) -> str:
        """
        Authorizes and returns a new auth token with entitlements.
        """
        auth_token = self._request('GET', '/admin/authorization')['token']
        return auth_token

    def authorize(self) -> None:
        while True:
            log.info('Authorizing...')
            try:
                self._auth_token = self._authorize()
                return
            except HTTPError as e:
                if e.response.status_code == 412:
                    # Precondition failed, requires admin OTP?
                    response = e.response.json()
                    log.info(response['message'])
                    otp_required = response.get('otpRequired', False)
                    if otp_required:
                        self.authenticate_otp()
                        continue

                message = e.response.json()['message']
                raise AdminError('Authorization failed: {}'.format(message)) from e

    def request(self, method: str, path: str, *, json: dict = None, headers: Dict[str, str] = None,
                files: Optional[Dict[str, Tuple[str, str, str]]] = None,
                timeout: float = DEFAULT_SOCKET_TIMEOUT, outfile: Optional[str] = None,
                progress_callback: Optional[ProgressCallback] = None,
                tries: Optional[int] = None) -> Dict[str, Any]:
        """
        Wrapper for _request that handles re-authentication and retries.

        :tries: Number of tries in case of generic server errors like connection errors,
         bad gateway or internal server errors. Set to 1 disable retries, None to use the default value.
        """
        # If unspecified, use the default value.
        if tries is None:
            tries = self._tries

        if tries < 1:
            log.warning('Tries should be greater or equal to 1, got %d, forcing to 1', tries)
            tries = 1

        # Generic retry logic
        for request_try in range(tries):
            try:
                # Specific retry logic for expired admin token
                for auth_try in range(2):
                    try:
                        return self._request(method, path, json=json, headers=headers, files=files,
                                             timeout=timeout, outfile=outfile, progress_callback=progress_callback)
                    except HTTPError as e:
                        if auth_try == 0 and e.response.status_code == 401 and \
                                self._username and \
                                self._provider and \
                                self._device_id:
                            # Token probably expired, re-authenticate!
                            log.warning('Admin token expired, re-authenticating')
                            self._auth_token = None
                            self.authenticate(username=self._username,
                                              provider=self._provider,
                                              device_id=self._device_id,
                                              password=self._password)
                            self.authorize()
                            continue
                        raise
            except (URLError, HTTPError) as e:
                # Out of tries
                if request_try >= (tries - 1):
                    raise

                # No point in retrying some HTTP errors
                if isinstance(e, HTTPError) and \
                        e.response.status_code in {400, 401, 404, 406, 409}:
                    raise

                # Retry after backoff time
                backoff_times = [3, 5, 10]
                backoff_time = min(backoff_times, key=lambda x: abs(x-request_try))
                if isinstance(e, socket.timeout):
                    log.warning('%s %s: Timed out. Retrying in %d seconds ...', method, path, backoff_time)
                else:
                    log.warning('%s %s: %s. Retrying in %d seconds ...', method, path, e, backoff_time)
                time.sleep(backoff_time)
        # Unreachable
        assert False

    def _check_file_repository_api(self) -> None:
        if self.peer_api_version < MINIMUM_VERSION_FILE_REPOSITORY_API:
            raise APIVersionError(self.peer_api_version, MINIMUM_VERSION_FILE_REPOSITORY_API)

    def file_upload(self, name: str, file: str, *, progress_callback: Optional[ProgressCallback] = None) -> None:
        self._check_file_repository_api()
        self.request('PUT', '/admin/files',
                     files={'file': (name, file, 'application/octet-stream')},
                     progress_callback=progress_callback)

    def file_start_download(self, name: str, url: str) -> None:
        self._check_file_repository_api()
        self.request('POST', '/admin/files',
                     json={'url': url, 'filename': name})

    def file_status(self, name: str) -> DetailedFileStatus:
        """
        The file status call can take a while as it verifies the checksum once the download is complete.
        """
        self._check_file_repository_api()
        r = self.request('GET', '/admin/files/{}'.format(name))
        return DetailedFileStatus(FileStatus(r['status']), r.get('checksum'), r.get('failureReason'))

    def file_delete(self, name: str) -> None:
        self._check_file_repository_api()
        self.request('DELETE', '/admin/files/{}'.format(name))
