import codecs
import fnmatch
import http.client
import time
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
import ipaddress
import re
from typing import Any, Optional, Dict, List, Set, NamedTuple, Tuple, Callable, Iterable, Sequence, Counter
import urllib.parse

from .admin import Admin, AdminError, HTTPError, URLError
from .logger import log

__all__ = [
    'MINIMUM_PEER_API_VERSION',
    'DEFAULT_THROTTLE',
    'DEFAULT_PREPARE_TIMEOUT',
    'DEFAULT_COMPLETE_TIMEOUT',
    'MINIMUM_PREPARE_TIMEOUT',
    'MINIMUM_COMPLETE_TIMEOUT',
    'TIMEOUTS',
    'ApplianceRole',
    'RoleStatus',
    'ALL_ROLES',
    'APPLIANCE_ROLE',
    'ALL_ROLES_STATUS',
    'MINIMUM_VERSION_HA_PATCH_CONTROLLER_UPGRADES',
    'UpgradeError',
    'UpgradeInProgressError',
    'Appliance',
    'APPLIANCE_SEMVER_REGEXP',
    'APPLIANCE_VERSION_REGEXP',
    'guess_target_version',
    'parse_semantic_version',
    'parse_server',
    'ApplianceAdmin',
    'ApplianceStatus',
    'ApplianceAttributeMatcher',
    'ApplianceFilter',
    'ApplianceMatcher',
    'ApplianceCombinedMatcher',
    'get_appliances_status',
    'log_status_changes',
    'build_appliance_filter',
    'UpgradeStatus',
    'DetailedUpgradeStatus',
    'APPLIANCE_FILTER_ATTRIBUTES',
    'AppliancesStartingStatus',
    'names',
    'wait_for_appliances_state',
    'wait_for_appliance_upgrade',
    'get_appliances_upgrade_status',
    'appliance_group_description',
]

# We need peer API version 7 to support the stateful upgrade API
MINIMUM_PEER_API_VERSION = 7

# How many appliances will download the upgrade image at a given time
DEFAULT_THROTTLE = 5

# How long to wait for an appliance to prepare for upgrade.
# This includes the time taken to download the upgrade file and verify its signature.
MINIMUM_PREPARE_TIMEOUT = 900.0
DEFAULT_PREPARE_TIMEOUT = 1800.0

# How long to wait for an appliance to complete an upgrade.
# This is the time it takes to install the upgrade to disk, reboot
# and complete data migration.
MINIMUM_COMPLETE_TIMEOUT = 300.0
DEFAULT_COMPLETE_TIMEOUT = 1200.0

# How long we wait for an appliance to cancel an upgrade.
DEFAULT_CANCEL_TIMEOUT = 600.0


# Use class variables to be able to set from another module
class TIMEOUTS:
    prepare = DEFAULT_PREPARE_TIMEOUT
    complete = DEFAULT_COMPLETE_TIMEOUT
    cancel = DEFAULT_CANCEL_TIMEOUT


class ApplianceRole(Enum):
    """
    Keep roles sorted by importance.
    """
    CONTROLLER = 'controller'
    LOGSERVER = 'logServer'
    LOG_FORWARDER = 'logForwarder'
    GATEWAY = 'gateway'
    CONNECTOR = 'connector'

    @property
    def description(self) -> str:
        """
        Return a user friendly name for the role.
        """
        return {
            ApplianceRole.CONTROLLER: 'controller',
            ApplianceRole.LOGSERVER: 'log server',
            ApplianceRole.LOG_FORWARDER: 'log forwarder',
            ApplianceRole.GATEWAY: 'gateway',
            ApplianceRole.CONNECTOR: 'connector',
        }[self]

    @property
    def importance(self) -> int:
        """
        Give the importance of the role, for example the controller role is always
        more important than the gateway role.
        This is useful when displaying information about an appliance to not clutter
        the output with roles. An appliance that is both a controller and a gateway
        can be called a controller.
        """
        for i, role in enumerate(ApplianceRole.__members__.values()):
            if self == role:
                return len(ApplianceRole.__members__) - i
        # Unreachable
        assert False

    def __lt__(self, other: 'ApplianceRole') -> bool:
        """
        Allows ordering roles by importance.
        """
        return self.importance < other.importance


def role_description_with_count(role: Optional[ApplianceRole], count: int) -> str:
    """
    Return role description, pluralize if count is greater than 1.
    For example:
        role=ApplianceRole.GATEWAY, count=3 -> 'gateways'
        role=ApplianceRole.GATEWAY, count=1 -> 'gateway'
        role=None, count=1 -> 'plain appliance'
    """
    return (
        role.description
        if role is not None else
        'plain appliance'
    ) + ('s' if count > 1 else '')


def appliance_group_description(appliances: List['Appliance']) -> str:
    """
    Return a description string for a group of appliances.
    For example: 'gateways, connector and plain appliances'
    """
    role_count = Counter[Optional[ApplianceRole]]()
    for appliance in appliances:
        if appliance.roles:
            role_count.update(appliance.roles)
        else:
            role_count.update((None,))

    # List roles sorted by importance along with their count.
    role_count_list = sorted(role_count.items(),
                             key=lambda i: i[0].importance if i[0] is not None else -1,
                             reverse=True)

    if len(role_count_list) > 1:
        role_str = ' and '.join((
            ', '.join(
                role_description_with_count(role, count)
                for role, count in role_count_list[:-1]
            ),
            role_description_with_count(*role_count_list[-1])
        ))
    else:
        role_str = role_description_with_count(*role_count_list[0])
    return role_str


ALL_ROLES = {role for role in ApplianceRole}


APPLIANCE_ROLE = 'appliance'
ALL_ROLES_STATUS = {role.value for role in ApplianceRole} | {APPLIANCE_ROLE}

# Patch controller upgrades with HA were added in peer API version 9
MINIMUM_VERSION_HA_PATCH_CONTROLLER_UPGRADES = 9


class UpgradeError(AdminError):
    pass


class UpgradeInProgressError(UpgradeError):
    def __init__(self, appliance_id: str, message: str) -> None:
        super().__init__(message)
        self.appliance_id = appliance_id


class Appliance:
    """
    Structure to represent an appliance with just the info we need to manage
    the upgrades.
    """
    __slots__ = (
        'appliance_id',
        'name',
        'peer_hostname',
        'peer_port',
        'admin_hostname',
        'admin_port',
        'roles',
        'version',
        'site',
        'log_forwarder_sites',
        'tags',
    )

    def __init__(self, appliance_id: str, name: str, peer_hostname: str, peer_port: int,
                 admin_hostname: Optional[str], admin_port: Optional[int],
                 roles: Set[ApplianceRole], version: int, site: Optional[str],
                 tags: List[str],
                 log_forwarder_sites: Optional[List[str]] = None) -> None:
        self.appliance_id = appliance_id
        self.name = name
        self.peer_hostname = peer_hostname
        self.peer_port = peer_port
        self.admin_hostname = admin_hostname
        self.admin_port = admin_port
        self.roles = roles
        self.version = version
        self.site = site
        self.log_forwarder_sites = log_forwarder_sites or []
        self.tags = tags

    def __repr__(self) -> str:
        params = []
        for field in Appliance.__slots__:
            params.append('%s=%s' % (field, repr(getattr(self, field))))

        return '{}({})'.format(self.__class__.__name__, ', '.join(params))

    def __eq__(self, other) -> bool:
        return type(other) == type(self) and self.appliance_id == other.appliance_id \
               and self.name == other.name and self.peer_hostname == other.peer_hostname \
               and self.roles == other.roles and self.site == other.site \
               and self.version == other.version

    def __hash__(self) -> int:
        return hash(self.appliance_id)

    @property
    def actual_admin_hostname(self) -> str:
        return self.admin_hostname or self.peer_hostname

    @property
    def actual_admin_port(self) -> int:
        return self.admin_port or self.peer_port

    @staticmethod
    def from_dict(data: dict) -> 'Appliance':
        version = data['version']
        roles = set()  # type: Set[ApplianceRole]
        for role in ALL_ROLES:
            role_name = role.value
            # Backwards compatibility with peer API version < 13
            # iotConnector was renamed to connector.
            if role == ApplianceRole.CONNECTOR and version < 13:
                role_name = 'iotConnector'
            if role_name in data and data[role_name]['enabled']:
                roles.add(role)

        return Appliance(
            appliance_id=data['id'],
            name=data['name'],
            peer_hostname=data['peerInterface']['hostname'],
            peer_port=data['peerInterface']['httpsPort'],
            admin_hostname=data['adminInterface']['hostname'] if 'adminInterface' in data else None,
            admin_port=data['adminInterface']['httpsPort'] if 'adminInterface' in data else None,
            version=version,
            roles=roles,
            site=data.get('site'),
            log_forwarder_sites=data['logForwarder']['sites'] if 'logForwarder' in data else [],
            tags=data['tags'],
        )


class RoleStatus:
    def __init__(self, name: str, status: str, details: Optional[str]) -> None:
        self.name = name
        self.status = status
        self.details = details

    @staticmethod
    def from_dict(role: str, data: dict) -> 'RoleStatus':
        return RoleStatus(
            name=role,
            status=data['status'],
            details=data.get('details')
        )


APPLIANCE_SEMVER_REGEXP = re.compile(r'^.*(\d+[.]\d+[.]\d+).*$')
APPLIANCE_VERSION_REGEXP = re.compile(r'^.*(\d+[.]\d+[.]\d+)-(\d+)-(\w+).*$')


def guess_target_version(image_url: str) -> Optional[str]:
    url = urllib.parse.urlparse(image_url)
    filename = url.path.rsplit('/', 1)[-1]

    # First try to match a version in the X.Y.Z-1234-branch form.
    m = APPLIANCE_VERSION_REGEXP.match(filename)
    if m:
        return '{semver}-{buildnumber}-{branch}'.format(semver=m.group(1),
                                                        buildnumber=m.group(2),
                                                        branch=m.group(3))

    # Then try to match a version in the X.Y.Z form.
    m = APPLIANCE_SEMVER_REGEXP.match(filename)
    if m:
        return m.group(1)

    # No guess...
    return None


def parse_semantic_version(version_string: str) -> Optional[str]:
    """
    Return semver version number from version string.
    Example:
        3.2.1-6220-release -> 3.2.1
    """
    match = APPLIANCE_SEMVER_REGEXP.match(version_string)
    if match:
        return match.group(1)
    else:
        return None


class UpgradeStatus(Enum):
    IDLE = 'idle'
    STARTED = 'started'
    DOWNLOADING = 'downloading'
    VERIFYING = 'verifying'
    READY = 'ready'
    INSTALLING = 'installing'
    SUCCESS = 'success'
    FAILED = 'failed'


DetailedUpgradeStatus = NamedTuple('DetailedUpgradeStatus', [
    ('status', UpgradeStatus),
    ('details', Optional[str])
])


class ApplianceStatus:
    def __init__(self, name: str, online: bool, state: Optional[str] = None,
                 version: Optional[str] = None, volume_number: Optional[int] = None,
                 roles: Optional[Dict[str, RoleStatus]] = None, upgrade: Optional[DetailedUpgradeStatus] = None) -> None:
        self.name = name
        self.online = online
        self.state = state
        self.version = version
        self.volume_number = volume_number
        self.semantic_version = parse_semantic_version(version) if version is not None else None
        if roles is not None:
            self.roles = roles
        else:
            self.roles = {}
        self.upgrade = upgrade

    def __eq__(self, other) -> bool:
        return isinstance(other, ApplianceStatus) and self.name == other.name \
               and self.online == other.online and self.state == other.state

    def __repr__(self) -> str:
        return '{}(name="{}", online={}, state="{}", version="{}", volume_number={}, roles={!r}, upgrade={!r})'.format(
            self.__class__.__name__, self.name, self.online,
            self.state or 'unknown', self.version or 'unknown',
            self.volume_number or 'unknown',
            self.roles,
            self.upgrade or 'unknown'
        )

    @property
    def appliance_busy(self) -> bool:
        return self.roles[APPLIANCE_ROLE].status == 'busy'

    @staticmethod
    def from_dict(data: dict) -> 'ApplianceStatus':
        upgrade = None  # type: Optional[DetailedUpgradeStatus]
        if 'upgrade' in data:
            upgrade = DetailedUpgradeStatus(status=UpgradeStatus(data['upgrade']['status']),
                                            details=data['upgrade'].get('details'))
        return ApplianceStatus(
            name=data['name'],
            online=data['online'],
            state=data.get('state'),
            version=data.get('version'),
            volume_number=data.get('volumeNumber'),
            roles={
                role: RoleStatus.from_dict(role=role, data=data[role])
                for role in ALL_ROLES_STATUS
                if role in data and data[role]['status'] != 'n/a'
            },
            upgrade=upgrade
        )


class ApplianceAdmin(Admin):
    def appliances(self) -> List[Appliance]:
        """
        Get all activated appliances.
        """
        appliances_data = self.request('GET', '/admin/appliances')['data']
        return [Appliance.from_dict(data)
                for data in appliances_data if data['activated']]

    def _set_controller_enabled(self, appliance_id: str, enabled: bool) -> None:
        data = self.request('GET', '/admin/appliances/{}'.format(appliance_id))
        data[ApplianceRole.CONTROLLER.value]['enabled'] = enabled
        self.request('PUT', '/admin/appliances/{}'.format(appliance_id), json=data)

    def disable_controller(self, appliance_id: str) -> None:
        self._set_controller_enabled(appliance_id, enabled=False)

    def enable_controller(self, appliance_id: str) -> None:
        self._set_controller_enabled(appliance_id, enabled=True)

    def appliances_status(self, *, tries: Optional[int] = None) -> List[ApplianceStatus]:
        """
        Example:
            [
                {
                    "controller": "healthy",
                    "cpu": 50.0,
                    "disk": 7.3,
                    "gateway": "n/a",
                    "id": "a9e38f05-c89d-4ae1-acb3-582f632b15fc",
                    "logging": "healthy",
                    "memory": 44.7,
                    "name": "controller-f0e2f8fb-80e4-40db-8902-fd9b7dc7238b-site1",
                    "online": true,
                    "version": "3.1.0-5240-testsd"
                },
                {
                    "controller": "healthy",
                    "cpu": 29.1,
                    "disk": 14.5,
                    "gateway": "n/a",
                    "id": "5138bf35-3d0e-4588-aa12-2f6cc780470c",
                    "logging": "n/a",
                    "memory": 13.4,
                    "name": "controller2-f0e2f8fb-80e4-40db-8902-fd9b7dc7238b-site1",
                    "online": true,
                    "version": "3.1.0-5240-testsd"
                }
            ]

        For peer API Versions 5 and higher the appliance state is also included.
        """
        return [ApplianceStatus.from_dict(data) for data
                in self.request('GET', '/admin/stats/appliances', tries=tries)['data']]

    def backup_begin(self, appliance_id: str, features: Dict[str, bool]) -> str:
        """
        Returns the ID of the backup operation
        """
        return self.request('POST', '/admin/appliances/%s/backup' % appliance_id, json=features)['id']

    def backup_status(self, appliance_id: str, backup_id: str) -> bool:
        """
        Returns true when the processing is complete
        """
        return self.request('GET', '/admin/appliances/%s/backup/%s/status' % (appliance_id, backup_id))['status'] == 'done'

    def backup_errors(self, appliance_id: str, backup_id: str) -> str:
        """
        Returns the stderr of the backup command on the appliance.

        This only returns meaningful data after the status has moved to 'done'. Prior
        only returns an empty string.
        """
        data = self.request('GET', '/admin/appliances/%s/backup/%s/status' % (appliance_id, backup_id))
        if 'output' in data:
            if data['encoding'] == 'utf-8':
                return data['output'].strip()
            else:
                # The data was not utf-8 so it was base64 encoded on the server.
                # Decoding it and returning a representation of it.
                return repr(codecs.decode(data['output'].encode('ascii'), 'base64'))
        return ''

    def backup_download(self, appliance_id: str, backup_id: str, destination: str) -> None:
        headers = {
            'Accept': 'application/vnd.appgate.peer-v%d+gpg' % self.peer_api_version
        }
        self.request(
            'GET',
            '/admin/appliances/%s/backup/%s' % (appliance_id, backup_id),
            headers=headers,
            outfile=destination
        )

    def backup_delete(self, appliance_id: str, backup_id: str) -> None:
        headers = {
            'Accept': 'application/vnd.appgate.peer-v%d+gpg' % self.peer_api_version
        }
        self.request(
            'DELETE',
            '/admin/appliances/%s/backup/%s' % (appliance_id, backup_id),
            headers=headers
        )

    def upgrade_appliance(self, appliance_id: str, image_url: str, dev_keyring: bool) -> None:
        json = {'imageUrl': image_url}  # type: Dict[str, Any]
        if self.peer_api_version >= 14:
            json['devKeyring'] = dev_keyring
        try:
            self.request('POST', '/admin/appliances/{}/upgrade'.format(appliance_id),
                         json=json)
        except HTTPError as e:
            if e.response.status_code == 409:
                raise UpgradeInProgressError(appliance_id, e.response.json()['message']) from None
            else:
                raise

    def upgrade_prepare(self, appliance_id: str, image_url: str, dev_keyring: bool) -> None:
        json = {'imageUrl': image_url}  # type: Dict[str, Any]
        if self.peer_api_version >= 14:
            json['devKeyring'] = dev_keyring
        try:
            self.request('POST', '/admin/appliances/{}/upgrade/prepare'.format(appliance_id),
                         json=json)
        except HTTPError as e:
            if e.response.status_code == 409:
                raise UpgradeInProgressError(appliance_id, e.response.json()['message']) from None
            else:
                raise

    def upgrade_complete(self, appliance_id: str, switch_partition: bool = False) -> None:
        body = {
            'switchPartition': switch_partition
        }
        try:
            self.request('POST', '/admin/appliances/{}/upgrade/complete'.format(appliance_id),
                         json=body)
        except HTTPError as e:
            if e.response.status_code == 409:
                raise UpgradeInProgressError(appliance_id, e.response.json()['message']) from None
            else:
                raise

    def upgrade_switch_partition(self, appliance_id: str) -> None:
        try:
            self.request('POST', '/admin/appliances/{}/upgrade/switchPartition'.format(appliance_id))
        except HTTPError as e:
            if e.response.status_code == 400:
                raise UpgradeError(e.response.json()['message']) from None
            else:
                raise

    def upgrade_status(self, appliance_id: str) -> DetailedUpgradeStatus:
        resp = self.request('GET', '/admin/appliances/{}/upgrade'.format(appliance_id))
        return DetailedUpgradeStatus(status=UpgradeStatus(resp['status']),
                                     details=resp.get('details'))

    def upgrade_cancel(self, appliance_id: str) -> None:
        try:
            self.request('DELETE', '/admin/appliances/{}/upgrade'.format(appliance_id))
        except HTTPError as e:
            if e.response.status_code == 400:
                raise UpgradeError(e.response.json()['message']) from None
            else:
                raise


def get_appliances_status(admin, appliance_names: Optional[List[str]] = None,
                          *, tries: Optional[int] = None) -> Dict[str, ApplianceStatus]:
    """
    Get the current status for the given list of appliance names (or all if unspecified).
    Return a dictionary of appliance names to appliance status.
    """
    appliances_status = admin.appliances_status(tries=tries)
    if appliance_names:
        filtered_status = (s for s in appliances_status if s.name in appliance_names)
    else:
        filtered_status = appliances_status
    return {s.name: s for s in filtered_status}


def log_status_changes(previous_appliances_status: Dict[str, ApplianceStatus],
                       current_appliances_status: Dict[str, ApplianceStatus]) -> None:
    """
    Log any changes in the appliance status in a human readable way.
    """
    for name, current in current_appliances_status.items():
        previous = previous_appliances_status.get(name)
        if previous != current:
            if previous is None:
                if current.online:
                    log.notice('Appliance %s is online and in state %s',
                               current.name, current.state if current.state else 'unknown')
                else:
                    log.info('Appliance %s is offline', current.name)
            elif previous.online != current.online:
                if current.online:
                    log.notice('Appliance %s is now online and in state %s',
                               current.name, current.state if current.state else 'unknown')
                else:
                    log.info('Appliance %s is now offline', current.name)
            elif previous.state != current.state:
                log.info('Appliance %s is now in state %s',
                         current.name, current.state if current.state else 'unknown')


def parse_server(server: str, port: int = 8443) -> Tuple[str, int]:
    """
    Allow server specification with or without the https scheme, with or without the port.
    Port from the controller takes precedence over the port argument
    """
    try:
        # Wrap an ipv6 ip with brackets in the URL.
        ip = ipaddress.ip_address(server)

        if ip.version == 6:
            server = '[{}]'.format(server)
    except ValueError:
        pass

    # Prepend https scheme if necessary
    has_scheme = server.find('://') > 0
    if not has_scheme:
        server = 'https://{}'.format(server)

    url = urllib.parse.urlparse(server)

    if not url.hostname:
        raise ValueError('Invalid hostname for controller: {}'.format(url.hostname))

    if url.scheme != 'https':
        raise ValueError('Invalid scheme for controller: {}'.format(url.scheme))

    try:
        if url.port:
            port = url.port
    except ValueError:
        raise ValueError('Invalid port for controller: {}'.format(server)) from ValueError

    return url.hostname, port


ApplianceFilter = Callable[[Iterable[Appliance]], Iterable[Appliance]]


class ApplianceMatcher:
    def __call__(self, appliance: Appliance) -> bool:
        return True


APPLIANCE_FILTER_ATTRIBUTES = {'role', 'function', 'id', 'site', 'name', 'tag'}


class ApplianceAttributeMatcher(ApplianceMatcher):
    """
    Returns an appliance matcher based on a match string.
    Valid match string are of the form key=value, value supports the glob format.
    """
    def __init__(self, match_str: str) -> None:
        try:
            attribute, expression = match_str.split('=', 1)
        except ValueError:
            raise UpgradeError('Invalid filter "{}", should be in attribute=value form, '
                               'with attribute one of {}'.format(
                                    match_str, ', '.join(sorted(APPLIANCE_FILTER_ATTRIBUTES))))

        if attribute not in APPLIANCE_FILTER_ATTRIBUTES:
            raise UpgradeError('Invalid filter attribute "{}", should be one of: {}'.format(
                attribute, ', '.join(sorted(APPLIANCE_FILTER_ATTRIBUTES))))

        self.attribute = attribute
        self.expression = expression

    def __repr__(self) -> str:
        return '{}(attribute="{}", expression="{}")'.format(self.__class__.__name__,
                                                            self.attribute,
                                                            self.expression)

    def __call__(self, appliance: Appliance) -> bool:
        attribute = self.attribute
        expression = self.expression
        if attribute in ('role', 'function'):
            return any(
                fnmatch.fnmatch(role.value, expression)
                for role in appliance.roles
            )
        elif attribute == 'id':
            return fnmatch.fnmatch(appliance.appliance_id, expression)
        elif attribute == 'site':
            return fnmatch.fnmatch(appliance.site or '', expression)
        elif attribute == 'name':
            return fnmatch.fnmatch(appliance.name, expression)
        elif attribute == 'tag':
            return any(
                fnmatch.fnmatch(tag, expression)
                for tag in appliance.tags
            )
        else:
            raise AttributeError(attribute)


class ApplianceCombinedMatcher(ApplianceMatcher):
    def __init__(self, matchers: Sequence[ApplianceMatcher]) -> None:
        self.matchers = matchers

    def __call__(self, appliance: Appliance) -> bool:
        return any(matcher(appliance) for matcher in self.matchers)


def _group_matchers_by_attribute(
        matchers: List[ApplianceAttributeMatcher]) -> Dict[str, List[ApplianceAttributeMatcher]]:
    """
    Group matchers by attribute
    """
    groups = {}  # type: Dict[str, List[ApplianceAttributeMatcher]]
    for matcher in matchers:
        if matcher.attribute not in groups:
            groups[matcher.attribute] = []
        groups[matcher.attribute].append(matcher)
    return groups


def build_appliance_filter(filters: Optional[List[str]] = None,
                           excludes: Optional[List[str]] = None) -> ApplianceFilter:
    """
    Filters a list of appliances based on one or more match strings.
    Valid match string are of the form key=value, value supports the glob format.

    If an appliance matches a single exclude it is ignored.
    If any filters are specified they are grouped by attribute.
    An appliance is ignored if it does not match any filter for the given attribute.
    """
    filter_matchers = list(map(ApplianceAttributeMatcher, filters or []))
    filters_by_attribute = _group_matchers_by_attribute(filter_matchers)
    filters_combined_matchers = [ApplianceCombinedMatcher(matchers=attribute_filters)
                                 for attribute_filters in filters_by_attribute.values()]

    exclude_matchers = list(map(ApplianceAttributeMatcher, excludes or []))

    def filter_fn(appliances: Iterable[Appliance]) -> Iterable[Appliance]:
        for appliance in appliances:
            # Any matching excludes excludes the appliance
            if exclude_matchers:
                if any(matcher(appliance) for matcher in exclude_matchers):
                    continue

            # Filters are grouped by attribute, all filter groups need to pass
            if filter_matchers:
                if not all(matcher(appliance) for matcher in filters_combined_matchers):
                    continue

            yield appliance

    return filter_fn


def names(appliances: Iterable[Appliance], sort: bool = True) -> str:
    """
    Returns the appliance names as a comma separated list of names.

    For example:
        [Appliance(name='foo', ...)] -> 'foo'
        [Appliance(name='foo', ...), Appliance(name='bar', ...)] -> 'foo, bar'
    """
    appliance_names = (appliance.name for appliance in appliances)
    return ', '.join(sorted(appliance_names) if sort else appliance_names)


def check_appliance_status(admin: Admin,
                           appliance_status: ApplianceStatus,
                           expected_state: str) -> Tuple[bool, str]:
    if not appliance_status.online:
        return False, 'appliance is offline'

    # We never consider an appliance to have reached the given state
    # if the status is still busy.
    if appliance_status.appliance_busy:
        return False, 'appliance is busy'

    # State and volume number were added to the appliance status in peer API version 5
    if admin.peer_api_version >= MINIMUM_PEER_API_VERSION:
        if appliance_status.state != expected_state:
            return False, 'expected state {}, got {}'.format(expected_state, appliance_status.state)

    return True, 'reached expected status'


def wait_for_appliances_state(admin: Admin,
                              appliances: List[Appliance], *,
                              num_attempts: int = 60,
                              seconds_between_attempts: int = 10,
                              expected_state: str = 'appliance_ready',
                              silent: bool = False) -> None:
    """
    Wait for the appliance to be online and in the expected state,
    defaulting to expecting appliance ready state.
    """
    appliance_names = [a.name for a in appliances]

    # Give some more time for log servers...
    if any(ApplianceRole.LOGSERVER in a.roles for a in appliances):
        # This will give it 20 minutes instead of 10 to sync the shards *and* do the upgrade
        num_attempts *= 2

    if not silent:
        log.info('Waiting for appliance%s to come back online in state %s: %s ...',
                 's' if len(appliance_names) > 1 else '',
                 expected_state, ', '.join(appliance_names))
    previous_appliances_status = {}  # type: Dict[str, ApplianceStatus]
    unreached_status = None  # type: Optional[List[Tuple[str, str]]]

    for _ in range(num_attempts):
        try:
            appliances_status = get_appliances_status(admin, appliance_names, tries=1)
            if not silent:
                log_status_changes(previous_appliances_status, appliances_status)
            previous_appliances_status = appliances_status

            # Check the status for all appliances
            checked_status = (
                (name, check_appliance_status(admin, appliance_status, expected_state))
                for name, appliance_status in appliances_status.items()
            )

            # Gather the list of appliance name that have not reached the status as well as the details
            unreached_status = [(name, details) for name, (reached, details) in checked_status if not reached]

            if not unreached_status:
                break
        except http.client.RemoteDisconnected:
            # Remote end closed connection without response,
            # this can happen when the master controller itself is upgraded.
            log.info('Controller closed connection without response, '
                     'the controller services are probably being restarted')
            pass
        except HTTPError as e:
            if e.response.status_code == 500:
                # Database is down for upgrade
                pass
            elif e.response.status_code == 502:
                # Services are restarting
                pass
            else:
                log.exception('Unhandled HTTP error while waiting for appliances')
        except URLError as e:
            # Appliance is restarting
            log.debug('Connection error while waiting for appliance: %s', e)
            pass
        except Exception:
            log.exception('Unhandled error while waiting for appliances')
        time.sleep(seconds_between_attempts)
    else:
        if unreached_status is not None:
            errors = ('Appliance {} never reached expected status: {}'.format(name, details)
                      for name, details in unreached_status)
            raise UpgradeError(', '.join(errors))
        else:
            raise UpgradeError('Failed to get appliances status')


VolumeNumbers = Dict[str, Optional[int]]


class AppliancesStartingStatus:
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance],
                 status: Optional[Dict[str, ApplianceStatus]] = None,
                 upgrade_status: Optional[Dict[str, DetailedUpgradeStatus]] = None) -> None:
        self.admin = admin
        self._appliances = appliances
        self._status = status
        self._upgrade_status = upgrade_status

    def initialize(self):
        if self._status is None:
            # Make sure the controllers are online and in the expected state
            controllers = [a for a in self._appliances if ApplianceRole.CONTROLLER in a.roles]
            wait_for_appliances_state(self.admin, controllers,
                                      num_attempts=1,
                                      silent=True,
                                      expected_state='multi_controller_ready'
                                      if len(controllers) > 1 else 'single_controller_ready')
            self._status = get_appliances_status(self.admin)

    @property
    def status(self) -> Dict[str, ApplianceStatus]:
        if self._status is None:
            raise UpgradeError('Appliance starting status not initialized!')
        return self._status

    @property
    def volume_numbers(self) -> VolumeNumbers:
        return {
            name: s.volume_number for name, s
            in self.status.items()
        }

    @property
    def versions(self) -> Dict[str, Optional[str]]:
        return {
            name: s.version for name, s
            in self.status.items()
        }

    @property
    def semantic_versions(self) -> Dict[str, Optional[str]]:
        return {
            name: s.semantic_version for name, s
            in self.status.items()
        }

    @property
    def upgrade_status(self) -> Dict[str, DetailedUpgradeStatus]:
        if self._upgrade_status is None:
            appliances_upgrade_status = get_appliances_upgrade_status(self.admin, self._appliances)
            self._upgrade_status = {
                appliance.appliance_id: upgrade_status
                for appliance, upgrade_status in appliances_upgrade_status
            }
        return self._upgrade_status

    def filter_appliances(self, appliances: Iterable[Appliance]) -> 'AppliancesStartingStatus':
        """
        Narrow down the appliance list.
        This is useful when excluding some appliances from the upgrade.
        """
        filtered_ids = set(appliance.appliance_id for appliance in appliances)
        filtered_names = set(appliance.name for appliance in appliances)
        appliances = [
            appliance for appliance in self._appliances
            if appliance.appliance_id in filtered_ids
        ]

        status = None  # type: Optional[Dict[str, ApplianceStatus]]
        if self._status is not None:
            status = {
                name: status for name, status in self._status.items()
                if name in filtered_names
            }

        upgrade_status = None  # type: Optional[Dict[str, DetailedUpgradeStatus]]
        if self._upgrade_status is not None:
            upgrade_status = {
                appliance_id: status for appliance_id, status in self._upgrade_status.items()
                if appliance_id in filtered_ids
            }

        return AppliancesStartingStatus(
            admin=self.admin,
            appliances=appliances,
            status=status,
            upgrade_status=upgrade_status
        )


def get_appliances_upgrade_status(admin: ApplianceAdmin,
                                  appliances: Iterable[Appliance], *,
                                  sequential: bool = False) -> Iterable[Tuple[Appliance, DetailedUpgradeStatus]]:
    """
    Returns an iterable of appliance and details upgrade status tuples.
    Can run either sequentially or concurrently, concurrently by default.
    """
    if sequential:
        for appliance in appliances:
            status = admin.upgrade_status(appliance.appliance_id)
            yield appliance, status
    else:
        def get_upgrade_status(a: Appliance) -> Tuple[Appliance, DetailedUpgradeStatus]:
            s = admin.upgrade_status(a.appliance_id)
            return a, s

        with ThreadPoolExecutor(max_workers=10) as executor:
            for appliance, upgrade_status in executor.map(get_upgrade_status, appliances):
                yield appliance, upgrade_status


def check_appliance_upgrade(appliance: Appliance,
                            description: str,
                            expected_state: str,
                            starting_volume_number: Optional[int],
                            previous_appliance_status: Optional[ApplianceStatus],
                            appliance_status: ApplianceStatus) -> Tuple[bool, ApplianceStatus]:
    title = description.capitalize()
    name = appliance.name

    if not appliance_status.online:
        # We shouldn't get here since an offline controller wouldn't be able to respond
        # to the status calls.
        if previous_appliance_status is None or previous_appliance_status.online:
            log.info('%s %s is offline', title, name)

        if previous_appliance_status and previous_appliance_status.online:
            # Don't override the whole status as we don't want to lose information
            # about previous volume number, state etc...
            previous_appliance_status.online = False
            return False, previous_appliance_status

        return False, appliance_status

    if appliance_status.state != expected_state:
        if previous_appliance_status is None or appliance_status.state != previous_appliance_status.state:
            log.info('%s %s is in state %s and volume number %s',
                     title, name, appliance_status.state, appliance_status.volume_number)
        return False, appliance_status

    if appliance_status.volume_number == starting_volume_number:
        if appliance_status.state == 'upgrading':
            # Still upgrading...
            return False, appliance_status

        upgrade_status = appliance_status.upgrade
        if upgrade_status:
            if upgrade_status.status == UpgradeStatus.FAILED:
                raise UpgradeError('Upgrade failed for {} {}: {}'.format(
                    description, name, upgrade_status.details))

            if upgrade_status.status == UpgradeStatus.IDLE:
                # The appliance isn't upgrading anymore, yet we are still on the same volume number.
                # This mean the upgrade was rolled back.
                raise UpgradeError('{} {} rolled back the upgrade!'.format(title, name))

        # Upgrade is still in progress...
        return False, appliance_status

    # All good!
    log.notice('%s %s has reached expected state %s and volume number %s',
               title, name, appliance_status.state, appliance_status.volume_number)
    return True, appliance_status


def wait_for_appliance_upgrade(admin: ApplianceAdmin,
                               appliances: List[Appliance],
                               description: str,
                               expected_state: str,
                               starting_volume_numbers: Dict[str, Optional[int]],
                               timeout: float) -> None:
    """
    Wait for the appliances to finish their upgrades and detect eventual rollbacks.
    We can't rely entirely on the controller API since it can be down during most of this phase.
    """
    assert timeout >= MINIMUM_COMPLETE_TIMEOUT
    seconds_between_attempts = 10
    num_attempts = int(timeout // seconds_between_attempts)

    start = time.monotonic()
    appliance_by_name = {
        a.name: a for a in appliances
    }

    log.info('Waiting for {}{} to come back online in state {} within %d seconds: {} ...'.format(
        description, 's' if len(appliances) > 1 else '', expected_state, names(appliances)), timeout)

    previous_appliances_status = {}  # type: Dict[str, ApplianceStatus]
    for i in range(num_attempts):
        if i > 0:
            time.sleep(seconds_between_attempts)

        try:
            appliances_status = get_appliances_status(admin, list(appliance_by_name.keys()), tries=1)
        except Exception:
            # If the master controller is upgrading, which implies stopping services and rebooting
            # expect disturbances...
            for appliance in appliance_by_name.values():
                previous_appliance_status = previous_appliances_status.get(appliance.name)
                if previous_appliance_status and previous_appliance_status.online:
                    # We didn't get an actual status, just set it offline
                    previous_appliance_status.online = False
                    log.info('%s %s is offline', description.capitalize(), appliance.name)
                    previous_appliances_status[appliance.name].online = False
            continue

        for name, appliance_status in appliances_status.items():
            appliance = appliance_by_name[name]
            previous_appliance_status = previous_appliances_status.get(name)
            starting_volume_number = starting_volume_numbers[name]
            done, previous_appliance_status = check_appliance_upgrade(
                appliance=appliance,
                description=description,
                expected_state=expected_state,
                starting_volume_number=starting_volume_number,
                previous_appliance_status=previous_appliance_status,
                appliance_status=appliance_status)
            previous_appliances_status[name] = previous_appliance_status
            if done:
                del appliance_by_name[name]
                del previous_appliances_status[name]

        # All appliances have reached the expected state!
        if len(appliance_by_name) == 0:
            return

    # Out of retries and some appliances still haven't reached the expected state, time out!
    elapsed = int(time.monotonic() - start)
    raise UpgradeError('Upgrade timed out after {} seconds for {}: {}!'.format(
        elapsed, description, names(appliance_by_name.values())))
