import logging
import os
import sys
import traceback
from typing import Optional, Dict

__all__ = [
    'log',
    'set_log_level',
    'LOG_LEVEL_MAPPING',
    'pick_color_function',
]


try:
    import xtermcolor
    _xtermcolormodule = True
except ImportError:
    _xtermcolormodule = False

try:
    import termcolor
    _termcolormodule = True
except ImportError:
    _termcolormodule = False


def cprint(string, color: Optional[str] = None):
    print(string, flush=True)


def cprint_xtermcolor(string, color: Optional[str] = None):
    colors = {
        'red': 1,
        'blue': 4,
        'green': 10,
        'yellow': 11,
        'magenta': 5,
        'white': 15,
    }  # type: Dict[str, int]
    if color:
        print(xtermcolor.colorize(string, ansi=colors[color]), flush=True)
    else:
        print(string, flush=True)


def cprint_termcolor(string, color: Optional[str] = None):
    termcolor.cprint(string, color, flush=True)


def pick_color_function() -> None:
    global cprint

    if not os.isatty(1):
        return

    if sys.platform.startswith('win'):
        return

    if _xtermcolormodule:
        cprint = cprint_xtermcolor
    elif _termcolormodule:
        cprint = cprint_termcolor


LOG_LEVEL_MAPPING = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR
}


class ColorHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        colormap = {
            'CRITICAL': 'red',
            'ERROR': 'red',
            'WARNING': 'yellow',
            # 'INFO':
            'DEBUG': 'blue',
            'NOTICE': 'green',
            # NOTSET	0
        }
        if record.exc_info:
            traceback.print_tb(record.exc_info[2])
        cprint(self.format(record), color=colormap.get(record.levelname))


NOTICE = 25
logging._levelToName[NOTICE] = 'NOTICE'


class ExtraLogger(logging.RootLogger):
    def notice(self, msg, *args, **kwargs):
        self._log(NOTICE, msg, args, **kwargs)


log = ExtraLogger(logging.INFO)  # type: ignore
log_handler = ColorHandler()
log_formatter = logging.Formatter('%(asctime)s %(message)s',
                                  datefmt='%H:%M:%S')
log_handler.setFormatter(log_formatter)
log.addHandler(log_handler)
log.setLevel(logging.INFO)


def set_log_level(level: str):
    log.setLevel(LOG_LEVEL_MAPPING[level.lower()])
