import abc
import enum
import functools
import socket
import sys
import time
import urllib.parse
from pathlib import Path
from typing import List, Optional, Generator, NamedTuple, Dict, Tuple

from .utils import plural, ask_for_confirmation
from .logger import log
from .appliance import Appliance, ApplianceAdmin, wait_for_appliances_state, names, AppliancesStartingStatus, \
    wait_for_appliance_upgrade, DetailedUpgradeStatus, UpgradeError, UpgradeStatus, get_appliances_upgrade_status, \
    UpgradeInProgressError, TIMEOUTS, appliance_group_description
from .admin import APIVersionError, HTTPError, FileStatus, ProgressCallback

__all__ = [
    'UpgradeStep',
    '_BaseUpgradeAppliancesStep',
    'EnableExtraControllersStep',
    'WarnAboutExtraControllerStep',
    'CancelAppliancesUpgradeStep',
    'DisableExtraControllersStep',
    'InstallAppliancesUpgradeStep',
    'PrepareAppliancesUpgradeStep',
    'UpgradeAppliancesStep',
    'DownloadUpgradeImageToControllerStep',
    'UploadUpgradeImageToControllerStep',
    'DeleteUpgradeImageFromControllerStep',
    'LazyImageURL',
    'ImageSpecifier',
    'ImageFilePath',
    'ImageURL',
]


class ImageSpecifier(metaclass=abc.ABCMeta):
    pass


class ImageFilePath(ImageSpecifier):
    """
    A file path was given.
    """
    def __init__(self, image_path: Path) -> None:
        self.image_path = image_path

    def __repr__(self) -> str:
        return '{}(image_path="{}")'.format(self.__class__.__name__, self.image_path)


class ImageURL(ImageSpecifier):
    """
    An URL was given.
    """
    def __init__(self, image_url: str, host_on_controller: bool = False) -> None:
        self.image_url = image_url
        self.host_on_controller = host_on_controller

    def __repr__(self) -> str:
        return '{}(image_url={}, host_on_controller={})'.format(self.__class__.__name__,
                                                                repr(self.image_url),
                                                                repr(self.host_on_controller))


class LazyImageURL:
    """
    The actual URL can be determined during the upgrade steps.
    """
    def __init__(self, value: Optional[str] = None) -> None:
        self.value = value

    def __repr__(self) -> str:
        return '{}(value={})'.format(self.__class__.__name__, repr(self.value))

    @property
    def image_url(self) -> str:
        assert self.value
        return self.value


class GoalState(enum.Enum):
    STARTED = 'started'
    CANCELLED = 'cancelled'
    PAUSED = 'paused'
    DOWNLOADED = 'downloaded'
    PREPARED = 'prepared'
    COMPLETED = 'completed'
    DONE = 'done'


ApplianceUpgradeGenerator = Generator[
    # At every step we yield the expected upgrade status
    GoalState,
    # We don't expect anything back from the yield
    None,
    # Nothing to return really...
    None
]


# Holds all the state relative to this appliance's upgrade.
ApplianceUpgradeState = NamedTuple('ApplianceUpgradeState', [
    ('generator', ApplianceUpgradeGenerator),
    ('appliance', Appliance),
    ('detailed_status', DetailedUpgradeStatus),
    ('goal_state', GoalState),
])


class GoalResult(enum.Enum):
    NOT_REACHED = 'not reached'
    REACHED = 'reached'
    FAILED = 'failed'


@functools.lru_cache(maxsize=len(GoalState)*len(UpgradeStatus))
def check_goal_result(goal_state: GoalState, upgrade_status: UpgradeStatus) -> GoalResult:
    """
    Check whether an appliance has reached it's goal state.
    Since we have a known number of inputs we can cache the results.
    """
    if goal_state == GoalState.STARTED:
        return GoalResult.REACHED

    elif goal_state == GoalState.CANCELLED:
        if upgrade_status == UpgradeStatus.IDLE:
            return GoalResult.REACHED

    elif upgrade_status == UpgradeStatus.FAILED:
        # Can't fail when starting or cancelling
        return GoalResult.FAILED

    elif goal_state == GoalState.PAUSED:
        return GoalResult.REACHED

    elif goal_state == GoalState.DOWNLOADED:
        if upgrade_status in {UpgradeStatus.VERIFYING, UpgradeStatus.READY,
                              UpgradeStatus.INSTALLING, UpgradeStatus.SUCCESS}:
            return GoalResult.REACHED

    elif goal_state == GoalState.PREPARED:
        if upgrade_status in {UpgradeStatus.READY, UpgradeStatus.INSTALLING,
                              UpgradeStatus.SUCCESS}:
            return GoalResult.REACHED

    elif goal_state == GoalState.COMPLETED:
        if upgrade_status == UpgradeStatus.SUCCESS:
            return GoalResult.REACHED

    else:
        assert False, 'unreachable'

    return GoalResult.NOT_REACHED


def log_upgrade_status_change(state: ApplianceUpgradeState, detailed_status: DetailedUpgradeStatus):
    goal_result = check_goal_result(state.goal_state, detailed_status.status)

    # Reached expected status!
    if goal_result == GoalResult.REACHED:
        log.notice('Upgrade status for %s reached expected value %s: %s',
                   state.appliance.name, detailed_status.status.value, detailed_status.details or 'no details')

    # Failed!
    elif goal_result == GoalResult.FAILED:
        log.error('Upgrade failed for %s: %s',
                  state.appliance.name, detailed_status.details or 'no details')

    # status changed
    elif detailed_status != state.detailed_status.status:
        log.info('Upgrade status for %s changed to %s: %s',
                 state.appliance.name, detailed_status.status.value, detailed_status.details or 'no details')


def run_upgrade_generators(admin: ApplianceAdmin, upgrade_states: Dict[str, ApplianceUpgradeState], *,
                           timeout: Optional[float] = None) -> None:
    """
    Advance through the given appliance upgrade generator until all appliances reach their goal state or failure.
    """
    last_change = time.monotonic()
    failed_upgrades_state = {}  # type: Dict[str, ApplianceUpgradeState]

    # Now advance all the upgrades to completion or error
    while upgrade_states:
        # Advance all appliance upgrades that can be advanced
        state_updated = False
        state_updates = []  # type: List[Tuple[str, ApplianceUpgradeState]]
        for appliance_id, state in upgrade_states.items():
            goal_result = check_goal_result(state.goal_state, state.detailed_status.status)

            if goal_result == GoalResult.NOT_REACHED:
                continue

            try:
                if goal_result == GoalResult.FAILED:
                    # Inject failure as an exception into the generator
                    error = '{}: {}'.format(state.appliance.name, state.detailed_status.details or 'no details')
                    # NOTE: Type checker doesn't check generator.throw correctly
                    goal_state = state.generator.throw(UpgradeError, error)  # type: ignore[arg-type]

                elif goal_result == GoalResult.REACHED:
                    goal_state = next(state.generator)

                else:
                    assert False, 'unreachable'

                if goal_state == state.goal_state:
                    # Goal state unchanged
                    continue

                log.debug('Goal state for %s changed to %s', state.appliance.name, goal_state.value)
                new_state = state._replace(goal_state=goal_state)
                state_updates.append((appliance_id, new_state))
            except StopIteration:
                new_state = state._replace(goal_state=GoalState.DONE)
                state_updates.append((appliance_id, new_state))
            except Exception as e:
                if not isinstance(e, UpgradeError):
                    detailed_status = state.detailed_status
                else:
                    detailed_status = DetailedUpgradeStatus(UpgradeStatus.FAILED, str(e))
                    log_upgrade_status_change(state, detailed_status)
                new_state = state._replace(goal_state=GoalState.DONE, detailed_status=detailed_status)
                state_updates.append((appliance_id, new_state))
                failed_upgrades_state[appliance_id] = new_state

        # Apply upgrade state updates
        for appliance_id, state in state_updates:
            if state.goal_state == GoalState.DONE:
                # Remove upgrade states that have reached their final state
                del upgrade_states[appliance_id]
            else:
                upgrade_states[appliance_id] = state
            state_updated = True

        # Exit early if done
        if not upgrade_states:
            break

        # Update the upgrade status for the remaining appliances
        remaining_appliances = (state.appliance for state in upgrade_states.values())
        try:
            for appliance, detailed_upgrade_status in get_appliances_upgrade_status(admin, remaining_appliances):
                state = upgrade_states[appliance.appliance_id]
                if detailed_upgrade_status != state.detailed_status:
                    # Log any upgrade status change
                    log_upgrade_status_change(state, detailed_upgrade_status)
                    upgrade_states[appliance.appliance_id] = state._replace(detailed_status=detailed_upgrade_status)
                    state_updated = True
        except socket.timeout:
            log.warning('Timeout while getting appliance upgrade status, retrying.')

        # Timeout if too much time elapsed since the last change
        if timeout:
            elapsed_since_last_change = time.monotonic() - last_change
            if elapsed_since_last_change > timeout:
                leftover_appliances = [state.appliance for state in upgrade_states.values()]
                raise UpgradeError('Upgrade timed out after {} seconds for appliance{}: {}'.format(
                    int(timeout), 's' if len(leftover_appliances) > 1 else '', names(leftover_appliances)
                ))

        if not state_updated:
            # Nothing happened, sleep a bit and wait for next iteration
            time.sleep(10.0)

    # Raise an error if any upgrade failed
    if failed_upgrades_state:
        errors = ['{}: {}'.format(state.appliance.name, state.detailed_status.details or 'no details')
                  for appliance_id, state in failed_upgrades_state.items()]
        raise UpgradeError(', '.join(errors))


class Throttler:
    """
    Holds the state relative to the throttling of prepared upgrades.
    """
    def __init__(self, throttle: int) -> None:
        self.throttle = throttle
        self.num_downloading = 0


class PrepareUpgradeState:
    """
    Holds the state relative to the prepare phase of a single appliance upgrade.
    If passed a throttler, the prepare upgrade will throttle downloads appropriately.
    """
    def __init__(self, admin: ApplianceAdmin, appliance: Appliance,
                 image_url: str, detailed_upgrade_status: DetailedUpgradeStatus,
                 throttler: Optional[Throttler], dev_keyring: bool) -> None:
        self.admin = admin
        self.appliance = appliance
        self.image_url = image_url
        self.throttler = throttler
        self.dev_keyring = dev_keyring
        self.detailed_upgrade_status = detailed_upgrade_status

    def generator(self) -> ApplianceUpgradeGenerator:
        if self.detailed_upgrade_status.status in {UpgradeStatus.READY, UpgradeStatus.SUCCESS}:
            # Upgrade was already prepared
            log.warning('Upgrade already prepared for %s', self.appliance.name)
            return

        if self.throttler:
            while True:
                if self.throttler.num_downloading < self.throttler.throttle:
                    break
                log.debug('Throttling download for %s, there are already %d out of %d appliances downloading',
                          self.appliance.name, self.throttler.num_downloading, self.throttler.throttle)
                yield GoalState.PAUSED

        if self.throttler:
            self.throttler.num_downloading += 1
            log.debug('Registering download in progress %d/%d',
                      self.throttler.num_downloading, self.throttler.throttle)

        try:
            try:
                log.info('Preparing upgrade for %s', self.appliance.name)
                self.admin.upgrade_prepare(self.appliance.appliance_id, self.image_url, self.dev_keyring)
                yield GoalState.DOWNLOADED
            except APIVersionError as e:
                raise UpgradeError('Prepared upgrades not supported '
                                   '(peer API version is {}, minimum is {})'.format(e.actual, e.minimum)) from None
            finally:
                if self.throttler:
                    self.throttler.num_downloading -= 1
                    log.debug('Unregistering download in progress %d/%d',
                              self.throttler.num_downloading, self.throttler.throttle)

            yield GoalState.PREPARED
        except Exception:
            log.debug('Cancelling upgrade for %s', self.appliance.name)
            self.admin.upgrade_cancel(self.appliance.appliance_id)
            raise


def prepare_appliances_upgrade(admin: ApplianceAdmin, appliances: List[Appliance], image_url: str, *,
                               throttle: Optional[int], dev_keyring: bool) -> None:
    throttler = None  # type: Optional[Throttler]
    if throttle:
        throttler = Throttler(throttle)

    def initialize_prepare_generator(appliance: Appliance,
                                     detailed_upgrade_status: DetailedUpgradeStatus) -> ApplianceUpgradeState:
        prepare_upgrade_state = PrepareUpgradeState(admin, appliance, image_url, detailed_upgrade_status,
                                                    throttler, dev_keyring)

        return ApplianceUpgradeState(
            generator=prepare_upgrade_state.generator(),
            appliance=appliance,
            detailed_status=detailed_upgrade_status,
            goal_state=GoalState.STARTED,
        )

    appliances_upgrade_status = get_appliances_upgrade_status(admin, appliances)
    upgrade_states = {
        appliance.appliance_id: initialize_prepare_generator(appliance, detailed_upgrade_status)
        for appliance, detailed_upgrade_status in appliances_upgrade_status
    }  # type: Dict[str, ApplianceUpgradeState]

    # Now advance all the upgrades to completion or error
    run_upgrade_generators(admin, upgrade_states, timeout=TIMEOUTS.prepare)


def cancel_appliances_upgrade(admin: ApplianceAdmin, appliances: List[Appliance]) -> None:
    def initialize_cancel_generator(appliance: Appliance,
                                    detailed_upgrade_status: DetailedUpgradeStatus) -> ApplianceUpgradeState:
        def generator(appliance: Appliance) -> ApplianceUpgradeGenerator:
            log.info('Cancelling prepared upgrade for %s', appliance.name)
            admin.upgrade_cancel(appliance.appliance_id)
            yield GoalState.CANCELLED

        return ApplianceUpgradeState(
            generator=generator(appliance),
            appliance=appliance,
            detailed_status=detailed_upgrade_status,
            goal_state=GoalState.STARTED,
        )

    appliances_upgrade_status = get_appliances_upgrade_status(admin, appliances)
    upgrade_states = {
        appliance.appliance_id: initialize_cancel_generator(appliance, detailed_upgrade_status)
        for appliance, detailed_upgrade_status in appliances_upgrade_status
    }  # type: Dict[str, ApplianceUpgradeState]

    # Now advance all the upgrades to completion or error
    run_upgrade_generators(admin, upgrade_states, timeout=TIMEOUTS.cancel)


def upgrade_appliance(admin: ApplianceAdmin,
                      appliance: Appliance,
                      detailed_upgrade_status: DetailedUpgradeStatus,
                      image_url: Optional[str],
                      dev_keyring: bool,
                      check_install_status: bool = True,
                      throttler: Optional[Throttler] = None,
                      ) -> ApplianceUpgradeGenerator:
    """
    Generator for a single appliance upgrade.
    Yield after each step is completed
    """
    # Full upgrade, start by preparing the appliance
    if image_url is not None:
        yield from PrepareUpgradeState(admin, appliance, image_url, detailed_upgrade_status,
                                       throttler, dev_keyring).generator()

    # Complete upgrade
    if check_install_status:
        if detailed_upgrade_status.status == UpgradeStatus.SUCCESS:
            # Upgrade was already installed, we just need to switch partition
            log.warning('Upgrade installation already complete for appliance %s', appliance.name)
        else:
            log.info('Completing upgrade on %s', appliance.name)
            try:
                admin.upgrade_complete(appliance.appliance_id)
            except UpgradeInProgressError:
                raise UpgradeError('Upgrade already in progress for appliance %s', appliance.name)
            yield GoalState.COMPLETED

        # Finalize upgrade by switching to the updated partition
        log.info('Switching partition on %s', appliance.name)
        try:
            admin.upgrade_switch_partition(appliance.appliance_id)
        except socket.timeout:
            log.warning('Response timed out when switching partition on %s, the appliance is probably rebooting',
                        appliance.name)
    else:
        # For the master controller we need to issue install and switch partition in one go since the database will
        # be brought down for the duration of the installation, reboot and data migration.
        log.info('Completing upgrade and switching partition on %s', appliance.name)
        try:
            admin.upgrade_complete(appliance.appliance_id, switch_partition=True)
        except UpgradeInProgressError:
            raise UpgradeError('Upgrade already in progress for appliance {}'.format(appliance.name))
        except socket.timeout:
            log.warning('Response timed out when switching partition on %s, the appliance is probably rebooting',
                        appliance.name)


def upgrade_appliances(admin: ApplianceAdmin, appliances: List[Appliance], image_url: Optional[str],
                       dev_keyring: bool,
                       check_install_status: bool = True, throttle: Optional[int] = None) -> None:
    appliances_upgrade_status = list(get_appliances_upgrade_status(admin, appliances))

    # Install prepared upgrade
    if image_url is None:
        # Verify that all appliances are ready for upgrade
        appliances_upgrade_not_ready = [(appliance, upgrade_status)
                                        for appliance, upgrade_status in appliances_upgrade_status
                                        if upgrade_status.status not in {UpgradeStatus.READY, UpgradeStatus.SUCCESS}]
        if appliances_upgrade_not_ready:
            error = '{} appliance{} not ready for upgrade: {}'.format(
                len(appliances_upgrade_not_ready),
                plural(appliances_upgrade_not_ready),
                ', '.join('{} ({})'.format(appliance.name, upgrade_status.status.value)
                          for appliance, upgrade_status in appliances_upgrade_not_ready)
            )
            raise UpgradeError(error)

    throttler = None  # type: Optional[Throttler]
    if throttle:
        throttler = Throttler(throttle)

    def initialize_upgrade_generator(appliance: Appliance,
                                     detailed_upgrade_status: DetailedUpgradeStatus) -> ApplianceUpgradeState:
        generator = upgrade_appliance(admin, appliance, detailed_upgrade_status,
                                      image_url, dev_keyring, check_install_status, throttler=throttler)
        return ApplianceUpgradeState(
            generator=generator,
            appliance=appliance,
            detailed_status=detailed_upgrade_status,
            goal_state=GoalState.STARTED,
        )

    upgrade_states = {
        appliance.appliance_id: initialize_upgrade_generator(appliance, detailed_upgrade_status)
        for appliance, detailed_upgrade_status in appliances_upgrade_status
    }  # type: Dict[str, ApplianceUpgradeState]

    # Now advance all the upgrades to completion or error
    run_upgrade_generators(admin, upgrade_states, timeout=TIMEOUTS.prepare)


class UpgradeStep:
    def __init__(self, admin: ApplianceAdmin) -> None:
        self.admin = admin

    @abc.abstractmethod
    def __call__(self) -> None:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def description(self) -> str:
        raise NotImplementedError()

    def rollback(self) -> Optional['UpgradeStep']:
        """
        Rollback a step.
        """
        return None


class EnableExtraControllersStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, master_controller: Optional[Appliance],
                 extra_controllers: List[Appliance]) -> None:
        super().__init__(admin=admin)
        self.master_controller = master_controller
        self.extra_controllers = extra_controllers

    @property
    def controllers(self) -> List[Appliance]:
        return ([self.master_controller] if self.master_controller else []) + self.extra_controllers

    @property
    def description(self) -> str:
        return 'Enable controller role on the following appliance{}: {}'.format(
            's' if len(self.extra_controllers) > 1 else '',
            ', '.join(a.name for a in self.extra_controllers))

    def __call__(self) -> None:
        # Enable the extra controllers one by one.
        # Wait for their states to stabilize to multi_controller_ready before
        # adding the next one.
        for i, controller in enumerate(self.extra_controllers, start=1):
            log.info('Enabling controller role on %s...', controller.name)
            self.admin.enable_controller(controller.appliance_id)

            if self.master_controller:
                controllers_ready = [self.master_controller]
            else:
                controllers_ready = []
            controllers_ready.extend(self.extra_controllers[:i])

            wait_for_appliances_state(self.admin, controllers_ready,
                                      seconds_between_attempts=10,
                                      num_attempts=360,
                                      expected_state='multi_controller_ready')


class WarnAboutExtraControllerStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, extra_controllers: List[Appliance]) -> None:
        super().__init__(admin=admin)
        self.extra_controllers = extra_controllers

    @property
    def description(self) -> str:
        return 'The controller role is still disabled on the following appliance{}: {}'.format(
            plural(self.extra_controllers), names(self.extra_controllers))

    def __call__(self) -> None:
        # If the controller upgrade failed when coming from a multi-controller configuration,
        # warn about manual rollback steps that can be taken.
        log.warning(self.description)
        log.warning('No more action will be taken in order to allow troubleshooting of the upgrade failure.')
        log.warning('You can rollback the upgrade on individual appliances by '
                    'connecting to them using SSH and running the following command:')
        log.warning('    sudo cz-config switch-partition --reboot')
        log.warning('Note that a controller rollback requires rolling back the upgrade on any other appliance that '
                    'has already been upgraded.')


class DisableExtraControllersStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, master_controller: Optional[Appliance],
                 extra_controllers: List[Appliance]) -> None:
        super().__init__(admin=admin)
        self.master_controller = master_controller
        self.extra_controllers = extra_controllers

    @property
    def description(self) -> str:
        return 'Disable controller role on the following appliance{}: {}'.format(
            's' if len(self.extra_controllers) > 1 else '',
            ', '.join(a.name for a in self.extra_controllers))

    def __call__(self) -> None:
        # Disable the extra controllers starting from the last, we will
        # enable them starting from the first.
        extra_controllers = list(reversed(self.extra_controllers))

        # Wait for the extra controllers to be disabled
        for i, controller in enumerate(extra_controllers, start=1):
            log.info('Disabling controller role on %s...', controller.name)
            self.admin.disable_controller(controller.appliance_id)
            wait_for_appliances_state(self.admin, [controller],
                                      expected_state='appliance_ready')

            if self.master_controller:
                controllers_ready = [self.master_controller]
            else:
                controllers_ready = []
            controllers_ready.extend(extra_controllers[i:])

            expected_state = 'multi_controller_ready' if len(controllers_ready) > 1 else 'single_controller_ready'
            wait_for_appliances_state(self.admin, controllers_ready,
                                      expected_state=expected_state)

    def rollback(self) -> Optional[UpgradeStep]:
        return WarnAboutExtraControllerStep(self.admin, self.extra_controllers)


def throttle_description(throttle: Optional[int], appliances: List[Appliance]) -> str:
    if throttle and len(appliances) > throttle:
        return ' {} at a time'.format(throttle) if throttle > 1 else ' sequentially'
    return ''


class PrepareAppliancesUpgradeStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance], image_url: LazyImageURL,
                 dev_keyring: bool,
                 throttle: Optional[int] = None) -> None:
        super().__init__(admin=admin)
        self.dev_keyring = dev_keyring
        self.appliances = appliances
        self.image_url = image_url
        assert throttle is None or throttle > 0
        self.throttle = throttle

    @property
    def description(self) -> str:
        return 'Prepare upgrade on the following appliance{plural}{throttle}: {names}'.format(
            plural='s' if len(self.appliances) > 1 else '',
            throttle=throttle_description(self.throttle, self.appliances),
            # Keep the given order when doing sequential upgrade preparation
            names=names(self.appliances, sort=not self.throttle == 1))

    def __call__(self) -> None:
        assert self.image_url.value
        prepare_appliances_upgrade(self.admin, self.appliances, self.image_url.value,
                                   throttle=self.throttle, dev_keyring=self.dev_keyring)


class CancelAppliancesUpgradeStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance]) -> None:
        super().__init__(admin=admin)
        self.appliances = appliances

    @property
    def description(self) -> str:
        return 'Cancel upgrade on the following appliance{}: {}'.format(
            's' if len(self.appliances) > 1 else '', names(self.appliances))

    def __call__(self) -> None:
        cancel_appliances_upgrade(self.admin, self.appliances)


class _BaseUpgradeAppliancesStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance], image_url: Optional[LazyImageURL],
                 starting_status: AppliancesStartingStatus,
                 appliance_description: Optional[str],
                 dev_keyring: bool,
                 expected_state: Optional[str] = None,
                 check_status: bool = True,
                 throttle: Optional[int] = None) -> None:
        super().__init__(admin=admin)
        self.dev_keyring = dev_keyring
        self.starting_status = starting_status
        self.appliances = appliances
        self.image_url = image_url
        self.expected_state = expected_state if expected_state is not None else 'appliance_ready'
        if appliance_description:
            self.appliance_description = appliance_description
        else:
            self.appliance_description = appliance_group_description(self.appliances)
        self.check_status = check_status
        # None means no throttle
        assert throttle is None or throttle > 0
        self.throttle = throttle

    @property
    def description(self) -> str:
        upgrade_command = 'Upgrade' if self.image_url is not None else 'Install prepared upgrade on'
        return '{upgrade_command} {description}{throttle}: {names}'.format(
            upgrade_command=upgrade_command,
            description=self.appliance_description,
            throttle=throttle_description(self.throttle, self.appliances),
            # Keep the given order when doing sequential upgrades
            names=names(self.appliances, sort=not self.throttle == 1))

    def __call__(self) -> None:
        image_url = self.image_url.value if self.image_url is not None else None
        upgrade_appliances(self.admin, self.appliances, image_url, self.dev_keyring,
                           check_install_status=self.check_status,
                           throttle=self.throttle)
        wait_for_appliance_upgrade(
            self.admin, self.appliances,
            expected_state=self.expected_state,
            description=self.appliance_description,
            starting_volume_numbers=self.starting_status.volume_numbers,
            timeout=TIMEOUTS.complete)


class UpgradeAppliancesStep(_BaseUpgradeAppliancesStep):
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance], image_url: LazyImageURL,
                 dev_keyring: bool,
                 starting_status: AppliancesStartingStatus, throttle: Optional[int] = None,
                 appliance_description: Optional[str] = None,
                 expected_state: Optional[str] = None,
                 check_status: bool = True) -> None:
        super().__init__(admin=admin, appliances=appliances, image_url=image_url, dev_keyring=dev_keyring,
                         starting_status=starting_status, throttle=throttle,
                         appliance_description=appliance_description, expected_state=expected_state,
                         check_status=check_status)


class InstallAppliancesUpgradeStep(_BaseUpgradeAppliancesStep):
    def __init__(self, admin: ApplianceAdmin, appliances: List[Appliance],
                 starting_status: AppliancesStartingStatus, throttle: Optional[int] = None,
                 appliance_description: Optional[str] = None,
                 expected_state: Optional[str] = None,
                 check_status: bool = True) -> None:
        super().__init__(admin=admin, appliances=appliances, image_url=None, dev_keyring=False,
                         starting_status=starting_status, throttle=throttle,
                         appliance_description=appliance_description, expected_state=expected_state,
                         check_status=check_status)


class DownloadUpgradeImageToControllerStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, source_image_url: str, image_url: LazyImageURL,
                 peer_hostname: str, peer_port: int, yes: bool = False) -> None:
        super().__init__(admin=admin)
        self.source_image_url = source_image_url
        self.image_url = image_url
        self.peer_hostname = peer_hostname
        self.peer_port = peer_port
        self.yes = yes
        # Assign a file name based on the URL
        url = urllib.parse.urlparse(self.source_image_url)
        self.name = url.path.rsplit('/', 1)[-1]

    @property
    def description(self) -> str:
        return 'Download upgrade image {} from source url {} to controller'.format(
            self.name, self.source_image_url)

    def __call__(self) -> None:
        # Start the download
        try:
            self.admin.file_start_download(self.name, self.source_image_url)
            log.info('Downloading upgrade image %s from source url %s to controller...',
                     self.name, self.source_image_url)
        except HTTPError as e:
            if e.response.status_code == 409:
                log.warning('File %s already exists, using it as is', self.name)
            else:
                raise

        # Wait for download to complete, give up after 15 minutes
        file_status = self.admin.file_status(self.name)
        for _ in range(90):
            if file_status.status == FileStatus.READY:
                break

            if file_status.status == FileStatus.FAILED:
                error = 'Failed to download file to controller'
                if file_status.failure_reason:
                    error += ': {}'.format(file_status.failure_reason)
                raise UpgradeError(error)

            # Download in progress, wait a bit
            time.sleep(10.0)

            file_status = self.admin.file_status(self.name)
        else:
            raise UpgradeError('Timeout while waiting for controller to download '
                               'file {} from source url {}'.format(self.name, self.source_image_url))

        # Assign controller based image URL
        self.image_url.value = 'controller://{}:{}/{}'.format(self.peer_hostname, self.peer_port, self.name)
        log.notice('Controller downloaded upgrade image %s%s, using image url %s',
                   self.name,
                   ' with sha256 checksum {}'.format(file_status.checksum) if file_status.checksum else '',
                   self.image_url.value)

    def rollback(self) -> Optional[UpgradeStep]:
        return DeleteUpgradeImageFromControllerStep(admin=self.admin, image_url=self.image_url,
                                                    yes=self.yes)


class UploadUpgradeImageToControllerStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, image_path: Path, image_url: LazyImageURL,
                 peer_hostname: str, peer_port: int, yes: bool = False) -> None:
        super().__init__(admin=admin)
        self.image_path = image_path
        self.image_url = image_url
        self.peer_hostname = peer_hostname
        self.peer_port = peer_port
        self.yes = yes
        self._progress = 0

    @property
    def description(self) -> str:
        return 'Upload upgrade image {} to controller'.format(self.image_path)

    def update_progress(self, bytes_sent: int, total: int) -> None:
        self._progress += bytes_sent
        progress_percent = (self._progress * 100) // total
        print('Uploaded {}%'.format(progress_percent), end='\r', flush=True)

    def check_file(self, name: str) -> bool:
        """
        Return True if the file is already there, False if it needs uploading.
        """
        try:
            file_status = self.admin.file_status(name)

            if file_status.status == FileStatus.FAILED:
                # File present but in failed state, time to re-upload it
                log.warning('File %s already exists%s, but is in status %s, overriding it', name,
                            ' with sha256 checksum {}'.format(file_status.checksum) if file_status.checksum else '',
                            file_status.status.value)
                return False

            elif file_status.status == FileStatus.READY:
                log.warning('File %s already exists%s, using it as is', name,
                            ' with sha256 checksum {}'.format(file_status.checksum) if file_status.checksum else '')
                return True

            else:
                raise UpgradeError('File {} already exists but is in status {}'.format(name, file_status.status.value))
        except APIVersionError as e:
            raise UpgradeError('Controller hosting of upgrade file is not supported '
                               '(peer API version is {}, minimum required is {})'.format(e.actual, e.minimum)) from None
        except HTTPError as e:
            if e.response.status_code == 404:
                # File not there, time to upload it
                return False
            else:
                raise

    def upload(self, name: str) -> None:
        try:
            # Print progress percentage if stdout is attached to a tty-like device
            progress_callback = None  # type: Optional[ProgressCallback]
            if sys.stdout.isatty():
                progress_callback = self.update_progress
            self.admin.file_upload(name, str(self.image_path), progress_callback=progress_callback)
        except HTTPError as e:
            if e.response.status_code == 409:
                log.warning('File %s already exists, using it as is', name)
            else:
                raise

    def __call__(self) -> None:
        name = self.image_path.name

        # Upload the appliance upgrade image file to the controller if necessary.
        if not self.check_file(name):
            self.upload(name)

        self.image_url.value = 'controller://{}:{}/{}'.format(self.peer_hostname, self.peer_port, name)
        log.notice('File %s uploaded to controller, using image url %s', name, self.image_url.value)

    def rollback(self) -> Optional[UpgradeStep]:
        return DeleteUpgradeImageFromControllerStep(admin=self.admin, image_url=self.image_url,
                                                    yes=self.yes)


class DeleteUpgradeImageFromControllerStep(UpgradeStep):
    def __init__(self, admin: ApplianceAdmin, image_url: LazyImageURL,
                 yes: bool = True) -> None:
        super().__init__(admin)
        self.image_url = image_url
        self.yes = yes

    @property
    def name(self) -> str:
        assert self.image_url.value
        name_ = urllib.parse.urlparse(self.image_url.value).path
        if name_[0] == '/':
            name_ = name_[1:]
        return name_

    @property
    def description(self) -> str:
        if self.image_url.value is None:
            return 'Delete upgrade image from controller'
        else:
            return 'Delete upgrade image {} from controller'.format(self.name)

    def __call__(self) -> None:
        if not self.image_url.value:
            log.warning('Nothing to delete')
            return

        name_ = self.name
        if not ask_for_confirmation(self.description, self.yes):
            log.info('Skipped')
            return

        log.info('Deleting appliance upgrade image file %s from controller', name_)
        try:
            self.admin.file_delete(name_)
        except HTTPError as e:
            # No need to rollback the whole upgrade if this fails so just log an error
            if e.response.status_code == 404:
                log.warning('Appliance upgrade image file %s not found on controller', name_)
            else:
                log.error('Failed to delete appliance upgrade image file %s from controller: %s', name_, e)
        except Exception:
            log.exception('Failed to delete appliance upgrade image file %s from controller', name_)
        log.notice('File %s deleted from controller', name_)
