from pathlib import Path
from typing import Sized
import uuid

__all__ = [
    'plural',
    'ask_for_confirmation',
    'default_device_id',
]


def plural(seq: Sized) -> str:
    """
    Return the plural or singular form depending on the number of elements to use when
    building a sentence.

    For example:
        f'The carrot{(plural(carrots)} blue

    would give, for 1 carrot:
        The carrot is blue

    for several carrots:
        The carrots are blue

    Disclaimer: does not work with sheep!
    """
    return 's are' if len(seq) > 1 else ' is'


def ask_for_confirmation(question: str, yes: bool = False) -> bool:
    """
    Returns a boolean based on the answer to the given question.
    """
    if not yes:
        while True:
            answer = input(question + ' [y/N]? ').strip().lower()
            if answer in {'y', 'yes'}:
                break
            elif answer in {'n', 'no', ''}:
                return False
    return True


def default_device_id() -> str:
    """
    Read default device id from machine id.
    Generate new uuid if the machine id could not be read.
    """
    try:
        machine_id = Path('/etc/machine-id').read_text().rstrip()
        # Re-format machine id as a device id
        return str(uuid.UUID(machine_id))
    except (FileNotFoundError, PermissionError):
        return str(uuid.uuid4())
